import os
import re
import sys
from pathlib import Path
import pytest
from sxpyr._exports import *
from sxpyr.walks import WalkRkt, WalkCl, WalkEl
from sxpyr.sxpyr import (
    conf_read, Walk, conf_plist, WalkPl, plist_to_dict, PList, Ast,
    make_do_path,
    UnknownStateError, DispatchNotImplementedError)

git_path = Path('~/git/').expanduser()
git_nofork_path = git_path / 'NOFORK'
test_data = Path(__file__).parent / 'data'

parse_common = configure()
parse_plist = configure(**conf_plist)
parse_sxpyr = configure(**conf_sxpyr)
parse_union = configure(**conf_union)
parse_cl = configure(**conf_cl)
parse_el = configure(**conf_el)
parse_xel = configure(**conf_xel)
parse_rkt = configure(**conf_rkt)
parse_gui = configure(**conf_gui)
parse_clj = configure(**conf_clj)
parse_hy = configure(**conf_hy)
parse_txr = configure(**conf_txr)


def sprint(gen, match=None, fail=False):
    res = []
    try:
        for parsed_expr in gen:
            if printexp:
                print('expr:', parsed_expr)

            res.append(parsed_expr)

        if fail:
            assert False, 'should have failed'
    except SyntaxError as e:
        if not fail:
            raise e
        else:
            print(f'task failed succssfully for ??? with {e}')
            return

    #print(res)
    if match is not None:
        assert res == match, (res, match)

    return res


def test_read():
    read_rkt = conf_read(parse_rkt, WalkRkt)
    read_el = conf_read(parse_el, WalkEl)

    # the dangers of recursion
    recursion_limit = sys.getrecursionlimit()
    rec_under = int(recursion_limit // 4.5)
    rec_over = int(recursion_limit // 4)
    print('under over on stack', rec_under, rec_over)
    # close enough to the limit that certain other things like calling
    # repr will hit the limit, however on versions of python newer than 3.6
    # it seems that repr no longer has this behavior (strange)
    big = next(read_rkt(''.join([p * rec_under for p in ("(", ")")])))
    try:
        repr(big)
        #assert False, "should have blown the stack"
    except RecursionError as e:
        pass

    try:
        oops = next(read_rkt(''.join([p * rec_over for p in ("(", ")")])))
        #assert False, "should have blown the stack"  # doesn't blow the stack in pypy3
    except RecursionError as e:
        pass

    # elisp chars
    a = next(read_el(r'?\^\M-\N{colon}'))
    b = next(read_el(r'?\^\M-:'))
    c = next(read_el(r'?\C-\M-:'))
    d = next(read_el(r'?\C-\M-\N{U+003a}'))
    assert a == b == c == d, f'{a} {b} {c} {d}'
    assert a.elc_int == 201326650, f'{a.elc_int} != 201326650'

    c1 = next(read_el(r'?\N{U+003A}'))
    c2 = next(read_el('?:'))
    assert c1 == c2, '{c1} != {c2}'

    # quasi-things and un-things
    sprint(read_rkt("#`(a #,b)"))
    sprint(read_rkt("#`(a #,@b)"))

    sprint(read_rkt("#'(a #,b)"))  # unq/s is allowed in quote or syntax
    sprint(read_rkt("#'(a #,@b)"))

    sprint(read_rkt("`(a #,b)"))
    sprint(read_rkt("`(a #,@b)"))

    sprint(read_rkt("#`(a ,b)"))
    sprint(read_rkt("#`(a ,@b)"))

    sprint(read_rkt("`(a ,@(b))"))
    sprint(read_rkt("`(a ,@(b `(c ,(d))))"))
    sprint(read_rkt("`(a ,@(b `(c ,(d `(e ,@(f))))))"))
    sprint(read_rkt("'(a ,@(b))"))
    sprint(read_rkt("`(a ,@(b ,(c)))"), fail=True)

    sprint(read_rkt("`(a ,(b))"))
    sprint(read_rkt("`(a ,(b `(c ,(d))))"))
    sprint(read_rkt("`(a ,(b `(c ,(d `(e ,(f))))))"))
    sprint(read_rkt("'(a ,(b))"))
    sprint(read_rkt("`(a ,(b ,(c)))"), fail=True)

    sprint(read_rkt("#`(a #,@(b))"))
    sprint(read_rkt("#`(a #,@(b #`(c #,(d))))"))
    sprint(read_rkt("#`(a #,@(b #`(c #,(d #`(e #,@(f))))))"))
    sprint(read_rkt("#'(a #,@(b))"))
    sprint(read_rkt("#`(a #,@(b #,(c)))"), fail=True)

    sprint(read_rkt("#`(a #,(b))"))
    sprint(read_rkt("#`(a #,(b #`(c #,(d))))"))
    sprint(read_rkt("#`(a #,(b #`(c #,(d #`(e #,(f))))))"))
    sprint(read_rkt("#'(a #,(b))"))
    sprint(read_rkt("#`(a #,(b #,(c)))"), fail=True)


def test_read_paths():
    def make_pr(parse, walk_cls):
        read = conf_read(parse, walk_cls)
        parse_path = make_do_path(parse)
        read_path = make_do_path(read)
        return read, parse_path, read_path

    def dol(pr, path):
        _, p, r = pr
        try:
            return list(p(path)), list(r(path))
        except RecursionError:
            pass
        except Exception as e:
            raise e

        raise RecursionError(f'exception in {path} {e}') #from e


    pr_common = read, parse_path, read_path = make_pr(parse_common, Walk)

    pr_pl = read_pl, parse_path_pl, read_path_pl = make_pr(parse_plist, WalkPl)  # FIXME WalkMax?

    pr_rkt = read_rkt, parse_path_rkt, read_path_rkt = make_pr(parse_rkt, WalkRkt)
    pr_el = read_el, parse_path_el, read_path_el = make_pr(parse_el, WalkEl)
    pr_cl = read_cl, parse_path_cl, read_path_cl = make_pr(parse_cl, WalkCl)

    pe1 = git_nofork_path / 'org-mode/lisp/ob-core.el'
    pe2 = git_nofork_path / 'emacs/test/src/syntax-tests.el'
    pe3 = git_nofork_path / 'emacs/lisp/simple.el' # anything but
    pr1 = git_nofork_path / 'racket/racket/collects/compiler/embed.rkt'
    pr2 = git_nofork_path / 'racket/racket/src/cs/schemified/expander.scm'  # big p

    p_rkt1, r_rkt1 = dol(pr_rkt, pr1)

    p_el1, r_el1 = dol(pr_el, pe1)
    p_el2, r_el2 = dol(pr_el, pe2)
    p_el3, r_el3 = dol(pr_el, pe3)

    p_rkt1, p_el1, p_el2, p_el3 = None, None, None, None  # debug repr slowness

    # test plist with caste to python
    ps1 = test_data / 'plist-test.sxpr'
    # (apply hash-map '(:a b :c d))
    # but need recursive
    # (apply hash-map '(:a b :c (:d e :f g)))
    p_1, r_1 = dol(pr_pl, ps1)
    def cf(ast):
        if isinstance(ast, PList):
            return plist_to_dict(ast.value)
        elif isinstance(ast, Ast):
            return ast.value
        else:
            return ast

    c_1 = [_.caste(cf) for _ in r_1]


def test_parse(debug=False):
    # asdf

    # racket wat
    sprint(parse_rkt("""
before
#|
        a
       /|\
      b c e
        |
        d
|#
after
    """))

    # hy
    # sprint(parse_hy("""#[[<{} object; :span (, 4 7) :match "bbb">]]"""))  # XXX TODO haven't implemented this

    # xemacs comma in comment in quasiquote
    sprint(parse_xel('`(a b c " d e, " f g\n)'))
    sprint(parse_xel('`(a b c ;; d e, f g\n)'))

    # plist sharp weirdness FIXME pretty sure that we don't want this
    # behavior where sharp eats until the next expression, or rather
    # it should eat, but internally Sharp ListP should be an error?
    sprint(parse_plist("""(
    # ; comment
    (sharpified thing))"""))

    # xelisp symbols and unquote in quasi
    sprint(parse_xel("`(a ,b,c)"))  # FIXME OH NO is it (unquote b) (unquote c) or what?!
    # in xemacs this reads with b,c as a single symbol in gemacs, it very much does not
    # and this is why you don't allow context sensitive things in the input layer

    sprint(parse_xel("`(a,b)"))
    # XXX so apparenly xemacs actually doesn't unquote b but instead
    # reads a,b as a symbol, gemacs however does what I would expect,
    # and I would classify the behavior here as a bug and require the
    # use of a\,b as in gemacs
    sprint(parse_xel("(`a,b)"))  # man there is a lot of weirdness in elisp

    # runnin (read "(list 1 , 2 , 3)") in gemacs well the explains a lot

    sprint(parse_xel("`(a ,b)"))
    sprint(parse_xel(r"`(a\,b)"))
    sprint(parse_xel("`(a ,(b `(c ,d)))"))
    sprint(parse_xel("`(1 ,(a,b `(3 ,c,d)))"))

    sprint(parse_xel("(',)"))
    sprint(parse_xel("`(,a)"))

    sprint(parse_xel("`,a"))
    sprint(parse_xel("`,\na"))
    # sprint(parse_xel("`,")) # FIXME this reads as nothing if using
    # C-x C-e however the form presented here does cause EOF errors in
    # nearly every dialect

    sprint(parse_xel('sigh-,-variants'), match=[Atom('sigh-,-variants')])

    # wat
    hrm = 'oh;no'
    sprint(parse_xel(hrm), match=[Atom('oh'), Comment('no')])
    hc_not_in_symbol = 'oh#*no'
    sprint(parse_sxpyr(hc_not_in_symbol), match=[Atom('oh#*no')])
    uok = '(;)\n)'
    sprint(parse_xel(uok), match=[ListP.from_elements(Comment(')'))])
    urg = '(x;)\n)'
    sprint(parse_xel(urg), match=[ListP.from_elements(Atom('x'), Comment(')'))])

    sprint(parse_cl('#"!"'))

    # FIXME cl and clojure double colon and beyond
    sprint(parse_cl('::k'))
    sprint(parse_cl(':::k'))
    sprint(parse_clj('::k'))
    sprint(parse_clj(':::k'))

    # FIXME racket ambiguous reading issues
    sprint(parse_rkt('#rx"lol"'))
    sprint(parse_rkt('#rx "lol"'))
    sprint(parse_rkt('#hash(a b c d)'))

    sprint(parse_rkt(""))
    sprint(parse_rkt("#{HRM}#"))  # FIXME ambiguous, used by guile
    sprint(parse_rkt("#{HRM} #"))  # with this

    sprint(parse_rkt("(# \"lol\")"))
    sprint(parse_rkt("#"))
    sprint(parse_rkt("##"))

    sprint(parse_rkt("'#:|fd\\sA|"))

    sprint(parse_clj("({'({:really? 'yes}) :wat 'thing :data} '({:really? 'yes}))"))
    sprint(parse_clj("({[{:really? 'yes}] :wat 'thing :data} [{:really? 'yes}])"))

    # these should read but will error later
    # because they have no label
    sprint(parse_rkt("#="))
    sprint(parse_rkt("##"))
    sprint(parse_rkt("#0=(1 2 3 . #0#)"))
    #sprint(parse_rkt("('##)"))  # TODO

    sprint(parse_rkt("#:|a b|"))
    # atom vs keyword distinction
    sprint(parse_rkt(":|a b|"))
    sprint(parse_rkt("|:a b|"))

    sprint(parse_rkt(":|d / e|"))
    sprint(parse_rkt("|:f / g|"))

    sprint(parse_rkt(r'#| #\; |#'))

    sprint(parse_rkt('#\\" lol'), match=[CharSpec('"'), Atom('lol')])
    sprint(parse_rkt('(#\\))'))
    sprint(parse_rkt(r'"."'))
    sprint(parse_rkt(r'#\0'))
    sprint(parse_rkt(r'#\.'))
    sprint(parse_rkt(r'#\space'))

    sprint(parse_rkt(r'"\""'))

    sprint(parse_rkt(r"(#\0)"))
    sprint(parse_rkt(r"#\0"))
    sprint(parse_rkt("\\#"))
    sprint(parse_rkt("'\\#"))

    # FIXME this means we have to preserve the escapes in atoms too!!!!
    sprint(parse_rkt('\1e+NaN')) # -> symbol
    sprint(parse_rkt('1e+NaN'))  # -> number

    # sharp things

    #sprint(parse_rkt("#"))
    sprint(parse_rkt("##HRM"))
    sprint(parse_rkt("#;1 2"))
    sprint(parse_rkt("#; 3 4"))
    sprint(parse_rkt("#; hello"), match=[XComment(Atom('hello'))])
    sprint(parse_hy("#_ there"), match=[XComment(Atom('there'))])
    sprint(parse_rkt("#a"))
    sprint(parse_rkt("#hash"))
    sprint(parse_rkt("#\\z"))
    sprint(parse_rkt("#'a"))
    sprint(parse_rkt("#`a"))
    sprint(parse_rkt("#,a"), match=[Sharp(UQuote(Atom('a')))])
    sprint(parse_rkt("#,@a"), match=[Sharp(SUQuote(Atom('a')))])

    # splicing unquote
    sprint(parse_rkt(r"'(regular , @ oops)"))
    sprint(parse_rkt(r"'(regular , @unquote)"))
    sprint(parse_rkt(r"'(splicing ,@unquote)"))
    sprint(parse_rkt(r"'(splicing ,@ unquote2)"))

    sprint(parse_el(r"(?\^?)"))

    sprint(parse_el(r'?a?b?-?\C-\C-\C-??d'))

    # weirdness with parsing chinese chars ... only happens if we read from file ??!?!
    # Oh great "冖" is really "AZ"" in byte from, of course emacs has support to auto convert
    # which is the file we were just looking at, but python hasn't he faintest idea
    # coding:iso-2022-7bit is in the file header which is why this was happening
    sprint(parse_el('(defvar lol \'(("冖")))'))

    # the ?\C-' issue
    sprint(parse_el(r"[?\C-\']"))

    # the ?\C-` issue
    sprint(parse_el(r"[?\C-\`]"))

    # the ?\C-? issue
    #sprint(parse_el(r"[?\C-?]"))  # XXX there is a note in the manual about this
    # stating that it is used all over the place ;_; grep -r "[?][\\]C-[?]"
    # not incredibly bad, but still problematic? and yes, requires special cases
    sprint(parse_el(r"[?\C-\?]"))  # sed -i "s/\([?][\\]C-\)[?]/\1\\\?/g"

    sprint(parse_el("(?\\123)"))  # oh no, python has this too ... implicit octal escapes
    # actually maybe it is a bug in what I'm doing here? hrm don't think so
    sprint(parse_el("(kp-separator ?,)"))
    # and best of all, of course you have to escape space, you just don't
    # have to escape other things that also end symbols, I can feel the
    # cruft of bad early design that as simply lingered for far too long
    sprint(parse_el("(define-key map [?\\S-\\ ] 'scroll-down-command)"))

    # FIXME FIXME sigh ?\C-, issue
    #sprint(parse_el("(define-key map [?\C-,]  'icomplete-backward-completions)"))  # FIXME now _THIS_ requires more states in the reader or tweaks to how m1 behaves specifically for C- and M- because ?\C- is its own thing and automatically escapes or goes back to state m1

    sprint(parse_el(r"(define-key map [?\C-\,]  'icomplete-backward-completions)"))  # this is how we can read such things, but full elisp compat requires the additional states as expected ... ARGH

    sprint(parse_el(r"(define-key map [?\M-\C-m] 'mpc-tagbrowser-dir-toggle)"))
    sprint(parse_el("(?\\N{lol I am not a unicode charachter name})"))  # FIXME this reads but incorrectly
    # NOTE we aren't going to fix this though since it is possible to recover
    sprint(parse_el("(?\\N{sigh}oops)"))  # oh boy, yeah, this is an error
    # but those curlies don't work like braces, they are just part of the symbol name
    # ugh what a mess, definitely not reading this stuff right now

    # charlits old
    sprint(parse_rkt(r'(list #\ a)'))
    sprint(parse_rkt(r'(list #\,a)'))
    sprint(parse_rkt(r'(list #\`a)'))
    sprint(parse_rkt(r'(list #\|a)'))

    sprint(parse_rkt(r'(list ?\ a)'))
    sprint(parse_rkt(r'(list ?\,a)'))
    sprint(parse_rkt(r'(list ?\`a)'))
    sprint(parse_rkt(r'(list ?\|a)'))

    sprint(parse_clj(r'(list \a\b\c)'))
    sprint(parse_clj(r'(list \a\b\c)'))
    # FIXME ?asdf and \asdf both cause read errors

    sprint(parse_rkt(r'(modify-syntax-entry ?\{  "(}1nb" table)'))
    sprint(parse_el(r'(modify-syntax-entry ?\{  "(}1nb" table)'))
    sprint(parse_el("?a?b?c"))
    sprint(parse_el("(list ?a?b?c)"))
    sprint(parse_el("""?
"""))

    sprint(parse_el("? "))
    sprint(parse_el("?)"))
    # wow, so this actually works, it is just hard to get
    # it to evaluate must be a legacy thing
    sprint(parse_el("?("))

    sprint(parse_rkt("#|| lol |||||#"))
    sprint(parse_rkt("a #| b #| c |# d |# e"))

    sprint(parse_rkt("(|Append| '|Append|)"))
    sprint(parse_cl("(`())"))
    #sprint(parse_cl("(a (b (c)) (d (e) `(f (g) h) ((i) j)))"))
    #sprint(parse_rkt("(a (b (c)) (d (e) `(f (g)  (1 (2)) (3 (4)) (ca 'q (* (f (1) (2))h))) ((i) j)))"))

    sprint(parse_rkt("a|b c|d"))
    sprint(parse_rkt("|a|b "))
    sprint(parse_rkt("|a| b"))

    # welcome to the watverse
    sprint(parse_rkt("(and 'a'b'c)"))

    sprint(parse_clj(":see-alsos [:clojure.core/* :clojure.core/*'],"))  # clojure weirdness  TODO

    # O ... K ... today we learned something about python strings
    # which brings us back to the question of why our example was failing
    sprint(parse_rkt(r";\
    "))
    sprint(parse_rkt(r";\
I can't believe you've done this."))
    sprint(parse_rkt(
        # yeah ... you just bashed yourself you idiot
        # \ escaping a literal newline is a rookie mistake
        # in a fit of extreme annoyance, there must be a newline between
        # the parens right now, maxiumum WAT
        # actually this is even more loltastic than you might realize
        # what I have /also/ done is to escape the newline at the end
        # of a comment so the comment eats the open paren on the newline
        # which is why I need _another_ newline in between the two parens
        # in order to evoke the error because you need the second newline
        # so that the closing parent isnt commented out

        # in common lisp you can in fact escape a newline but NOT in a comment
        # this is also true in racket and in elisp and they all protect against
        # the escape in the comment

        # XXX actually this is the result of the fact that inside a python string
        # you can escape a newline with a \ ... sigh
r""";\
( 
)"""
#"""
))
    sprint(parse_rkt('(#\\))'))
    sprint(parse_cl(
        # common lisp reader making our day super bad for square brackets
    r"""
(with-test (:name (compile * :constant-behind-cast :lp-1717971]))
  (checked-compile-and-assert (:optimize :quick/incomplete)
      `(lambda (x)
         (declare (type integer x))
         (declare (optimize (space 3) (compilation-speed 0)))
         (catch 'ct5
           (* (flet ((%f (&key (x (throw 'ct5 123)))
                       (the integer x)))
                (%f))
              x)))
    ((45) 123)))
"""))
    sprint(parse_rkt(r'#\['))
    sprint(parse_rkt('#| | |#'))

    sprint(parse_rkt(r'"\\"'))
    sprint(parse_rkt(r'("\\")'))
    sprint(parse_rkt(r'[("\\")]'))

    sprint(parse_rkt("(define lol #f)"))
    sprint(parse_rkt("[define hah #t]"))

    sprint(parse_rkt("{:a 1 :b 2 :c 3}"))
    sprint(parse_rkt("{:a 1 :b {hello there} :c 3}"))

    sprint(parse_rkt("#'stx"))
    sprint(parse_rkt("#`stx"))

    # feature expressions
    sprint(parse_rkt("#+\n()"))
    sprint(parse_rkt("#-\n()"))
    sprint(parse_rkt("#- (not feature)"))

    sprint(parse_rkt("#+()"))
    sprint(parse_rkt("#-()"))
    sprint(parse_rkt("#-(and yes I have features)"))
    sprint(parse_rkt("#-nil"))
    sprint(parse_rkt("#+nil"))
    sprint(parse_rkt("#+nil #-nil #-nil #-nil 'a"))
    sprint(parse_rkt("#+(or oh no) woop"))

    # eXpression comments
    sprint(parse_rkt("#;\n(a #; b c)"))
    sprint(parse_rkt("(a #; b c)"))
    sprint(parse_rkt("(a #_ b c)"))
    sprint(parse_union("(a #_#; b c)"))
    sprint(parse_rkt("(a #; #; b c)"))
    sprint(parse_union("(a #_ #_ b c)"))

    sprint(parse_rkt("' ; shi\nt"))

    sprint(parse_rkt('(a"b\\"c"d)'), match=[
        ListP.from_elements(
            Atom('a'),
            EString(['b', SEscape('"'), 'c']),
            Atom('d'),)
    ])  # should fail

    sprint(parse_rkt("('())"), match=[
        ListP.from_elements(Quote(ListP.from_elements()),)
    ])

    sprint(parse_rkt("('()) a"), match=[
        ListP.from_elements(Quote(ListP.from_elements()),),
        Atom('a')
    ])

    sprint(parse_rkt("(X\"\")"), match=[
        ListP.from_elements(
            Atom('X'),
            '')
    ])
    sprint(parse_rkt("'(1'q)"), match=[
        Quote(
            ListP.from_elements(
                Atom('1'),
                Quote(Atom('q'))))
    ])
    sprint(parse_rkt("'(('q)a)"), match=[
        Quote(
            ListP.from_elements(
                ListP.from_elements(
                    Quote(Atom('q')),),
                Atom('a')))
    ])

    sprint(parse_rkt("('(x))"))
    sprint(parse_rkt("()"))
    sprint(parse_rkt("(())"))
    sprint(parse_rkt("(()())"))
    sprint(parse_rkt("'(1)"))
    sprint(parse_rkt("'(1())"))
    sprint(parse_rkt("'(('q))"))
    sprint(parse_rkt("''''''''''q"))

    # quasiquote
    sprint(parse_rkt("(`(x))"))
    sprint(parse_rkt("`(1)"))
    sprint(parse_rkt("`(1())"))
    sprint(parse_rkt("`((`q))"))
    sprint(parse_rkt("``````````q"))
    sprint(parse_rkt("`(1(2)`(3)(`4)`(`5))"))

    sprint(parse_rkt("(a (b)) c"))
    sprint(parse_rkt("'(1(2)'(3)) "))
    sprint(parse_rkt("'(1(2)'(3))"))

    sprint(parse_rkt("'(1(2))"))
    sprint(parse_rkt("'(1(2)'3)"))
    sprint(parse_rkt("'(1(2)'(3)('4))"))
    sprint(parse_rkt("'(1(2)'(3)) oops"))
    sprint(parse_rkt("'(1(2)'(3)('4)f)"))
    sprint(parse_rkt("'(1(2)'(3)('4)())"))
    sprint(parse_rkt("'(1(2)'(3)('4)'())"))
    sprint(parse_rkt("'(1(2)'(3)('4)'('5))"))
    sprint(parse_rkt('(a (b))\nc'))
    sprint(parse_rkt('(d ; comment\n(e))\nf'))
    sprint(parse_rkt(';comment'))
    sprint(parse_rkt('(symbol :keyword 1234 "string")'))
    sprint(parse_rkt('symbol'))
    sprint(parse_rkt('1234'))
    sprint(parse_rkt('"string"'))
    sprint(parse_rkt('t'))
    sprint(parse_rkt('#t'))
    sprint(parse_rkt('#f'))
    sprint(parse_rkt("'()"))
    sprint(parse_rkt("'quoted-symbol"))
    sprint(parse_rkt("       leading whitespace"))
    sprint(parse_rkt("trailing whitespace               "))
    sprint(parse_rkt("   'leading"))
    sprint(parse_rkt("'trailing    "))
    sprint(parse_rkt("'(((((((((((())))))))))))"))
    sprint(parse_rkt("wat 'trailing"))
    sprint(parse_rkt("symbol'quote more"))
    sprint(parse_rkt(":keyword'quote"))
    sprint(parse_rkt("''quote"))
    sprint(parse_rkt("'quote'quote"))
    sprint(parse_rkt("'quote(sym1)"))
    sprint(parse_rkt("'quote[sym2]"))
    sprint(parse_rkt("sym(sym3)"))
    sprint(parse_rkt("sym[sym4]"))
    sprint(parse_rkt("1[sym5]"))
    sprint(parse_rkt("2[sym6]"))
    sprint(parse_rkt("(a (b))"))
    sprint(parse_rkt("'(a (b))"))
    sprint(parse_rkt("'(1 2 3 4 (asdf asdf))"))
    sprint(parse_rkt("'(5 6 7 8 (asdf asdf))\n(desire to know more intensifies)"))
    sprint(parse_rkt("(defun asdf (there) 'hello there)"))
    sprint(parse_rkt("I am multiple atoms that should be read in a row. (indeed) \"yes\""))
    sprint(parse_rkt('(testing "I am a happy string \\" yay!" string escape)'))
    # NOTE if you have unmatched parens in a here string that means that
    # this parser will fail
    sprint(parse_rkt("""
;;; hrm
#;
(
#<<-- oh boy
don't look now, a
here strings ... the absolute worst because
your reader is no longer stateless, unlikely
that these will be allowed, they induce quite
a bit for complexity
 oh boy
)
"""))

    test_cfg1 = '''(:config-search-paths
("user-1.yaml" "{:user-config-path}/program-name/config.yaml")
:auth-variables (
:var-whose-default-value-is-nil/null/None ; FIXME
:var-name-1 ""
)
)
'''
    sprint(parse_rkt(test_cfg1))


def test_chars(debug=False):
    # charachter literals
    def make_test_char(char_marker):
        def inner (s):
            if isinstance(s, str):
                return '(' + char_marker + s + ')'
            else:
                return '(' + ''.join([char_marker + ss for ss in s]) + ')'

        return inner

    test_suffixes = (
              ("\\", ListP.from_elements(CharSpec('\\'),)),
              (" ", ListP.from_elements(CharSpec(' '),)),
              (";", ListP.from_elements(CharSpec(';'),)),
              (")", ListP.from_elements(CharSpec(')'),)),
              ("z", ListP.from_elements(CharSpec('z'),)),  # can't use a because emacs -> x07
              ("  a", ListP.from_elements(CharSpec(' '), Atom('a'))),
              ("  'a", ListP.from_elements(CharSpec(' '), Quote(Atom('a')))),
              (") a", ListP.from_elements(CharSpec(')'), Atom('a'))),
              (" 'a", ListP.from_elements(CharSpec(' '), Quote(Atom('a')))),
              (")'a", ListP.from_elements(CharSpec(')'), Quote(Atom('a')))),
              (" a", ListP.from_elements(CharSpec(' '), Atom('a'))),
              (";a", ListP.from_elements(CharSpec(';'), Atom('a'))),
              (")a", ListP.from_elements(CharSpec(')'), Atom('a'))),
              ("'a", ListP.from_elements(CharSpec("'"), Atom('a'))),
              ("aa", ListP.from_elements(CharSpec('aa'),)),  # XXX REMINDER that non-existent char errors are NO LONGER reader errors
              (("y", "z"), ListP.from_elements(CharSpec('y'), CharSpec('z'))),
              ("newline", ListP.from_elements(CharSpec('newline'),)),
              #("SpAcE", ListP.from_elements(CharSpec('space'),)),  # TODO this only works after walk
              )
    effs = dict(
        tcq = make_test_char('?'),
        tcqe = make_test_char('?\\'),
        tcs = make_test_char('#\\'),
        tcb = make_test_char('\\'),
    )
    for key, fun in effs.items():
        for (string, ast) in test_suffixes:
            # todo for parser in (parse, parse_cl) etc. based on which fun/key is used
            tstr = fun(string)
            if debug:
                print('tc:', repr(tstr))
            if key in ('tcq', 'tcqe'):
                parser = parse_el
            elif key == 'tcb':
                parser = parse_clj
            else:
                parser = parse_rkt

            #print('string:', string)
            if (#string == 'aa' and key != 'tcqe' or
                #string == 'newline' and key == 'tcqe' or
                string == "'a" and parser == parse_clj):# or string == 'a)' and key == '':
                sprint(parser(tstr))  # will fail in the second pass
                #try:
                    #sprint(parser(tstr), match=[ast])
                    #assert False, 'should have failed'
                #except ValueError as e:
                    #pass
            elif (string == '\\' and key == 'tcq'):
                try:
                    sprint(parser(tstr), match=[ast])
                    assert False, 'should have failed'
                except SyntaxError as e:
                    pass
            elif key == 'tcqe':
                if ast is not None:
                    ast = ast.from_elements(*(ECharSpec([SEscape(_.value)])
                                              if isinstance(_, CharSpec)
                                              else _ for _ in ast.collect))

                # a number of these now read without error and the
                # fact that there is no such charachter is deferred
                if string == ' a':
                    ast = ListP.from_elements(ECharSpec([SEscape(' '), 'a']),)
                elif string == ';a':
                    ast = ListP.from_elements(ECharSpec([SEscape(';'), 'a']),)
                elif string == ')a':
                    ast = ListP.from_elements(ECharSpec([SEscape(')'), 'a']),)
                elif string == '\'a':
                    ast = ListP.from_elements(ECharSpec([SEscape('\''), 'a']),)
                elif string == 'aa':
                    ast = ListP.from_elements(ECharSpec([SEscape('a'), 'a']),)
                elif string == 'newline':
                    ast = ListP.from_elements(ECharSpec([SEscape('n'), *'ewline']),)

                sprint(parser(tstr), match=[ast])
            else:
                sprint(parser(tstr), match=[ast])


@pytest.mark.skipif('CI' in os.environ, reason='Need git paths and os paths.')
def test_parse_paths():
    paths = (
        #(git_nofork_path / 'racket/racket/share/pkgs/future-visualizer/future-visualizer/tests/visualizer.rkt'),
        (git_path / 'protc/anno-tags.rkt'),
        #(git_nofork_path / 'sbcl/src/compiler/generic/early-objdef.lisp'),
        #(git_nofork_path / 'sbcl/src/code/print.lisp'),
        #Path('/usr/share/guile/2.2/ice-9/i18n.scm'),
        #(git_nofork_path / 'emacs/lisp/international/titdic-cnv.el'),
        #(git_nofork_path / 'emacs/lisp/isearch.el'),
        #(git_nofork_path / 'emacs/lisp/icomplete.el'),
        #(git_nofork_path / 'emacs/lisp/mpc.el'),
        #(git_nofork_path / 'emacs/test/src/syntax-tests.el'),
        #(git_nofork_path / 'clojure/test/clojure/test_clojure/errors.clj'),
        #(git_nofork_path / 'racket/racket/src/cs/schemified/schemify.scm'),  # broken due to |#%name|
        #(git_nofork_path / 'racket/racket/src/schemify/wrap.rkt'),
        #(git_nofork_path / 'sbcl/tests/data/wonky3.lisp'),
        (git_nofork_path / 'hy/tests/native_tests/hy_repr.hy'),  # broken due to #[strings]

        Path('/usr/lib/xemacs/xemacs-packages/lisp/edebug/edebug.el'),
        Path('/usr/lib/xemacs/xemacs-packages/lisp/tm/latex-math-symbol.el'),
        Path('/usr/lib/xemacs/xemacs-packages/lisp/footnote/footnote-kana.el'),
        Path('/usr/lib/xemacs/xemacs-packages/lisp/haskell-mode/haskell-decl-scan.el'),
        Path('/usr/lib/xemacs/xemacs-packages/lisp/psgml/psgml-dtd.el'),

        #)
        #_sigh = (

        # txr
        *(git_nofork_path / 'txr').rglob('*.tl'),

        # fennel
        *(git_nofork_path / 'Fennel').rglob('*.fnl'),

        # edn
        #test_data / 'export.edn',
        *(git_nofork_path / 'edn-dot-net').rglob('*.edn'),

        # clj
        *(git_nofork_path / 'clojure').rglob('*.clj'),
        *(git_nofork_path / 'spec.alpha').rglob('*.clj'),
        *(git_nofork_path / 'core.specs.alpha').rglob('*.clj'),
        *(git_nofork_path / 'tawny-owl').rglob('*.clj'),

        # hy
        *Path('/usr/lib/python3.7/site-packages/hy/contrib').rglob('*.hy'),
        *(git_nofork_path / 'hy').rglob('*.hy'),

        # racket
        *(git_path / 'protc/protc-lib/protc').rglob('*.rkt'),
        *(git_nofork_path / 'racket').rglob('*.rkt'),  # wow ... fast

        # chez
        *(git_nofork_path / 'racket').rglob('*.scm'),

        # guile
        *Path('/usr/share/guile/2.2/').rglob('*.scm'),

        # gambit (wow that install location)
        *Path('/usr/lib64/').rglob('*.scm'),

        # common lisp
        *(git_nofork_path / 'sbcl').rglob('*.lisp'),
        *(git_nofork_path / 'ccl').rglob('*.lisp'),

        # elisp
        *(git_nofork_path / 'emacs/').rglob('*.el'),  # elisp uses ? for chars
        *(git_nofork_path / 'org-mode/').rglob('*.el'),  # elisp uses ? for chars
        # https://yoo2080.wordpress.com/2013/11/25/question-mark-and-emacs-lisp/

        # xemacs elisp
        *Path('/usr/lib/xemacs').rglob('*.el'),

             )

    should_fail = (
        'compiler-output-test.lisp',
        'wonky3.lisp',

        # racket + racket chez layer
        #'interpret.rkt',   # FIXME not really should fail but |asdf asdf| not impl
        #'infer-name.rkt',  # |[| issue
        'mk-gdbinit.rkt',  # here string killing things
        'rktc-utils.rkt',  # at-exp declared on 3rd line, comment on previous lines
        'new.rkt',         # here string
        #'xform.rkt',       # |;| issue
        #'schemify.scm',    # |#%name| issue
        'pat-unify.rkt',   # #lang 2d issues
        'chat-noir-literate.rkt',  # #lang scribble/lp
        'std-grammar.rkt', # sciribble issues
        #'scribble/reader.rkt',  # here string

        # guile
        # https://www.gnu.org/software/guile/manual/html_node/Symbol-Read-Syntax.html
        # section 6.6.6.6 appropriately enough, except no mention in guile about the 'a'b'c problem
        #'psyntax-pp.scm',  # #{}# guile has some weird #{ }# syntax thing ... what the heck is this?!
        #'psyntax.scm',     # #{}#
        #'i18n.scm',        # guile allows things like number' to be symbols ?!?!?!?!? apparently so
        #'parser.scm',      # #{}#
        #'compile-tree-il.scm',  # #{}# no idea what this is, but it is annoying, seems like the cl/rkt pipe?
        #'function-slot.scm',  # #{}# seems like it is being used to define elisp functions or something?
        #'autofrisk.scm',   # #{}#

        # elisp
        #'admin.el',        # ? issue
        #'unidata-gen.el',  # ?
        #'syntax-tests.el', # hah this is handy, and yeah the ? issue
        #'print-tests.el',  # elisp allows '## ?? what the heck is it even?
        #'json-tests.el',   # '##
        #'org-list.el',     # '##
        #'autoinsert.el',   # ?
        #'woman.el',        # ?
        #'minibuffer.el',   # ?
        #'completion.el',   # ?
        #'json.el',         # ?
        #'pcmpl-unix.el',   # ?
        #'skeleton.el',     # ?
        'titdic-cnv.el',    # coding:iso-2022-7bit kills this one

        #'simple.el',       # ?\C-? issue
        #'icomplete.el',    # ?\C-, issue
        #'ido.el',          # ?\C-, issue
        #'isearch.el',      # ?\^? issue
        #'iswitchb.el',     # ?\C-, issue
        #'iris-ansi.el',    # ?\C-' ?\C-` and ?\C-, issues
        #'ns-win.el',       # ok, I give up, elisp char syntax is too complex
        #'xterm.el',        # ?\C-, infuratingly the ?\C-\' on the line above it _was_ commented

        # xelisp
        'tex-jp.el',        # coding:iso-2022-jp-unix
        'mew-lang-jp.el',   # unmarked coding:iso-2022-7bit-unix emacs uses hueristics
        'riece-mcat-japanese.el',  # coding: iso-2022-jp
        'footnote-kana.el',        # coding system issues
        'latex-math-symbol.el',    # unmarked coding:iso-2022-7bit-unix emacs uses hueristics

        # clojure
        #'reducers.clj',    # *' issue
        #'transducers.clj', # end with ' issue ... need to look into this ... very rare to see but not banned
        #'memorise_test.clj',  # #{ issue
        #'owl.clj',         # end with '

        # hy
        'hy_repr.hy',  # #[string] issues
        'compiler_error.hy',  # expected unterminated string

        # fennel
        *(_.name for _ in (git_nofork_path / 'Fennel/test/bad/').rglob('*.fnl')),

        # txr
        'ffi-misc.tl',  # #b'string' issues?
        'split.tl', # #/regex/ issues?
        'termios.tl',  # `str\ning` issues
    )
    fail = True
    for count, p in enumerate(paths):
        p = p.expanduser()
        with open(p, 'rt') as f:
            try:
                hrm = f.read()
            except UnicodeDecodeError as e:
                print('fail unicode:', p)
                continue

            if hrm.startswith('#lang at-exp'):
                # can't parse @exp
                continue

            elif hrm.startswith('#lang scribble'):
                continue

            try:
                if p.suffix == '.tl':
                    tres = sprint(parse_txr(hrm))

                res = sprint(parse_rkt(hrm))  # racket reader compatible syntax
                print('success:', p)
                #val = list(parse_rkt(hrm))
            except Exception as err:
                try:
                    if p.name in should_fail:
                        continue
                    elif p.suffix == '.tl':
                        if re.search(r'`(?!`).*`', hrm):
                            print('txr fail: `string`')
                            continue
                        elif '#b' in hrm:
                            print('txr fail: #b binary notation')
                            continue
                        else:
                            raise err
                    elif p.suffix == '.rkt':
                        if '#<<' in hrm:
                            print('racket fail: here string', p)
                            # here string
                            continue
                    elif p.suffix == '.lisp':
                        print('generic failed:', p, err)
                        res = sprint(parse_cl(hrm))
                        print('success cl:', p)
                        #breakpoint()
                        # TODO
                        continue
                    elif p.suffix == '.el':
                        print('generic failed:', p, err)
                        try:
                            res = sprint(parse_el(hrm))
                            print('success el:', p)
                        except Exception as err3:
                            if 'xemacs' in p.as_posix():
                                print('fail el:', p)
                                res = sprint(parse_xel(hrm))
                                print('success xel:', p)
                            else:
                                raise err3 from err
                            #if '?' in hrm:
                                #print('fail ? issue:', p, err2)
                            #else:
                                #raise err2
                        continue
                    elif p.suffix in ('.clj', '.edn'):
                        print('generic failed:', p, err)
                        res = sprint(parse_clj(hrm))
                        print('success clj:', p)
                        continue
                    elif p.suffix == '.scm' and 'guile' in p.as_posix():
                        print('generic failed:', p, err)
                        res = sprint(parse_gui(hrm))
                        print('success gui:', p)
                        continue
                    elif p.suffix == '.hy':
                        print('generic failed:', p, err)
                        res = sprint(parse_hy(hrm))
                        print('success hy:', p)
                        continue

                except Exception as err2:
                    raise err2.__class__(p) from err2

                debug = True
                try:
                    sprint(parse_rkt(hrm))
                except:
                    pass

                debug = False

                # racket files that use |these types of ids|
                # or escape brackets |[| cause issues
                if fail:
                    raise SyntaxError(p) from err
                else:
                    print(err)
                    print('fail path:', p)

            #sprint(v)

    print('total paths:', count + 1)


def test_fails_read():
    read_rkt = conf_read(parse_rkt, WalkRkt)
    bads = ('#',
            '# ',
            '##',
            '#_',
            "#+ ()",
            # #\\\x1a is ok #\^Z -> #\u001A but not #^Z
            # this should fail in read but not parse
            # it triggers bad syntax in racket
            '#\x1a',
            )
    run_fails(read_rkt, bads)


def test_fails_parse():
    bads = (
        ',',
        "(defun fail (parse)",
        '"I am an unterminated string',
        'unmatched ) right paren',
        'unmatched ] right bracket',
        "{",
        "}",
        '"',
        "'",  # need a proper error
        "' ; oops",  # should fail but is not
        "[)",
        "(]",
        "'(1(2)')",  # confusing error message
        "|a| b|",
        '#| #| |#',
        '(;hrm',  # comment state could mask unterminated list
        '(#;',
        '(#||#',
        '#;',
        '#\\',  # note #\ unterminated char in racket
            )
    run_fails(parse_rkt, bads)


def run_fails(read_or_parse, bads):
    should_have_failed = []
    for bad in bads:
        try:  # I WANT MACROS
            sprint(read_or_parse(bad))
            should_have_failed.append(bad)
        except (UnknownStateError, DispatchNotImplementedError) as e:
            print(f'task failed succssfully as {e!r} for {bad!r}')
        except SyntaxError as e:
            print(f'task failed succssfully for {bad!r}')
        except NotImplementedError as e:
            print(f'task failed succssfully as not implemented for {bad!r}')

    assert not should_have_failed, should_have_failed


if __name__  == '__main__':
    debug = '-d' in sys.argv
    printexp = debug or '-e' in sys.argv
    if debug:
        import sxpyr.sxpyr
        sxpyr.sxpyr.debug = True

    test_read()
    test_read_paths()
    #sys.exit()
    test_parse(debug=debug)
    test_parse_paths()
    test_fails_read()
    test_fails_parse()
    test_chars(debug=debug)

else:
    printexp = False



import ast
import pathlib
from pyontutils.utils import asStr
here = pathlib.Path(__file__).parent

state_names_full = dict(
    bos= 'Top Level',
    s_atom= 'Atom',
    s_keyw= 'Keyword',
    s_quote= 'Quote',
    s_unquote= 'Unquote',
    s_unq_first= 'Unquote First',
    s_unq_splice= 'Unquote Splicing',
    s_quasiq= 'Quasiquote',
    s_string= 'String',
    s_list_p= 'List Paren',
    s_list_s= 'List Square',
    s_list_c= 'List Curly',
    s_sharp= 'Dispatch',  # dispatch
    s_comm= 'Comment',
    s_esc= 'Escape',
    s_esc_str= 'Escape String',
    s_comment_x= 'Comment Expression',
    s_comment_h= 'Comment Heading',  # no current dialects have this, inspired by org
    s_feat_x= 'Expression Feature',
    s_cblk= 'Comment Block',
    s_pi_in_cblk= 'Comment Block Maybe Exit',  # maybe leave block comment
    s_atom_verbatim= 'Atom Verbatim',
    s_comment_b_nest= 'Comment Block Nested',
    s_char= 'Charachter Literal',
    s_char_first= 'Charachter Literal First',
    )

token_names_abstract = dict(
    # common
    t_newline='Newline',
    t_tab='Tab',
    t_space='Space',
    t_beg_list_p='List Paren Begin',
    t_end_list_p='List Paren End',

    t_beg_end_str='String Delimiter',

    t_to_sharp='Dispatch',

    t_to_quote='Quote',
    t_to_quasi='Quasiquote',
    t_to_unquote='Unquote',

    t_to_splc_in_unq='Splice in Unquote',

    t_to_esc_in_str='Escape in String',

    t_to_comment='Comment',

    # uncommon
    t_beg_list_s='List Square Begin',
    t_end_list_s='List Square End',
    t_beg_list_c='List Curly Begin',
    t_end_list_c='List Curly End',

    t_beg_end_aver='Verbatim Atom Delimiter',

    t_to_cblk_in_shrp='Block Comment in Dispatch',
    t_to_xcom_in_shrp='Expression Comment in Dispatch',
    t_to_hcom_in_shrp='Heading Comment in Dispatch',  # XXX
    t_to_keyw_in_shrp='Keyword in Dispatch',
    t_to_feat_in_shrp='Feature Expression in Dispatch',
    t_to_strn_in_shrp='Expression Comment in Dispatch',  # TODO only used in hy, adds more states, not sure if want

    t_to_unquote_in_quasi='Unquote in Quasiquote',  # this is among the most evil of all

    t_to_keyw='Keyword',
    t_to_esc='Escape',

    t_to_char='Charachter',
)

def analysis():
    with open(here / 'sxpyr.py', 'rt') as f:
        tree = ast.parse(f.read())

    ast_conf = [n for n in tree.body if isinstance(n, ast.FunctionDef) and n.name == 'configure'][0]
    ast_state = [n for n in ast_conf.body if isinstance(n, ast.Assign) and n.targets[0].id == 'states'][0]

    conf_args = [arg.arg for arg in ast_conf.args.args]
    abstract_tokens = [t for t in conf_args if t.startswith('t_')]
    n_abstract_tokens = [token_names_abstract[t] for t in abstract_tokens]
    states = [n.id for n in ast_state.targets[-1].elts]
    n_states = [state_names_full[s] for s in states]
    return n_abstract_tokens, ['-------------'], n_states



# burn baby burn

from unicodedata import name as uniname, lookup as unilookup


class ELispCharThing:
    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return f'<ELCT {self.value!r}>'

    def __hash__(self):
        return hash((self.__class__, self.value))

    def __eq__(self, other):
        return type(self) == type(other) and self.value == other.value


class KeyChar:
    """ Base for char escape things. """


class UnicodeName(KeyChar):
    """ Elisp madness. """


class NeedDash(KeyChar):
    """ Sigh. """


class MetaNeedDash(NeedDash):
    """ Alt to the rest of us. """
    bit = 2 ** 27


class HyperNeedDash(NeedDash):
    """ ??? to the rest of us. """
    bit = 2 ** 24


class AltNeedDash(NeedDash):
    """ Alt to the rest of us.
        Not sure about relation to Meta here. """
    bit = 2 ** 22


class ControlHat(KeyChar):
    """ Maybe this is cmd on macs? No idea. """
    bit = 2 ** 26


class ControlNeedDash(NeedDash):
    """ Maybe this is cmd on macs? No idea. """
    bit = 2 ** 26


class ShiftNeedDash(NeedDash):
    r""" ?\S- ?? No idea. """
    bit = 2 ** 25


class SpaceOrSuperNeedDash(NeedDash):
    r""" ?\s- ?? No idea. """
    bit = 2 ** 23


class Octal(KeyChar):
    r""" ?\000 to ?\777 """

    def __init__(self, integer):
        self.value = integer

    def __repr__(self):
        return f'Octal({self.value})'


cee_base = {
    'n': '\n',
    't': '\t',
    'r': '\r',
    'v': '\v',
    'b': '\b',
    'a': '\a',
    'f': '\f',
}

cee_oct = {
    '0': Octal(0), # OH DEAR, these read as octal
    '1': Octal(1),
    '2': Octal(2),
    '3': Octal(3),
    '4': Octal(4),
    '5': Octal(5),
    '6': Octal(6),
    '7': Octal(7),
}

cee_el = {
    # OH NO
    # given its origins in Emacs, elisp has has by far the
    # largest number of ways to represent the same charachter
    # not all of these are implemented by this reader since
    # even reading chars at all induces way more complexity
    # than I originally wanted
    **cee_base,
    #**cee_oct,
    'd': '\x7f',  # delete aka rubout
    'e': '\x1b',  # escape
    's': SpaceOrSuperNeedDash,  # space
    #'s-': ????,  # FIXME AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
    'C': ControlNeedDash,  # FIXME must be followed by a dash
    'M': MetaNeedDash,     # and emacs denotes this with a -1
    'S': ShiftNeedDash,
    'H': HyperNeedDash,
    'A': AltNeedDash,
    'N': UnicodeName,  # must be followed by {...}
    '^': ControlHat,  # emacs denotes this with a -1
}

cee_gui = {
    **cee_base,
    #**cee_oct,
}


known_multi = {
    # from Racket, this is how racket manages to implement the
    # char lit reader, they have a closed set of valid multichar
    # charachters whereas cl has an arbitrary number because
    # syntactically it allows any symbol name and then fails if a
    # char by that name is not know to the implementation
    # except of course for all the \u unicode escapes
    # NOTE elisp is the only one that supports ?\x all the rest
    # only support \u or #\u, except hy, which is py which has no chars
    'backspace': '\x08',  # aka '\b'
    'newline': '\x0a',    # ala '\n'
    'linefeed': '\x0a',   # ala '\n'
    'nul': '\x00',
    'null': '\x00',       # "\u0000" "\x00"
    'page': '\x0c',       # \f
    'return': '\x0d',     # \r
    'rubout': '\x7f',     # aka delete
    'space': '\x20',
    'tab': '\x09',        # \t
    'vtab': '\x0b',       # \v
}


def dialect_char(collect):
    class self: """ lol """

    #assert len(collect) == 1 or , collect
    if len(collect) == 1:
        self.value = collect[0]
    elif not collect:
        raise ValueError('Empty char??')
    else:
        print('charco:', collect)
        # yes, let's just add the need to handle higher order
        # logics and errors when using what appears to be a simple
        # predicate ... LOL PYTHON

        #if (isinstance(collect[0], type) and issubclass(collect[0], KeyChar) or
            #isinstance(collect[0], Octal)):
        if isinstance(collect[0], SEscape):
            # EEEEEMMMMMAAAAAACCCCCSSSSSS
            self.value = ELispCharThing(collect)
            return

        if v in self.known_multi:  # FIXME sbcl unicode #\LONG_NAME_THINGS
            self.value = self.known_multi[v]
        elif v.lower() in self.known_multi:  # FIXME and downcasing reader thing
            self.value = self.known_multi[v.lower()]
        elif v.startswith('u') or v.startswith('U') or v.startswith('x'):
            # FIXME the racket reader works differently for x
            self.value = chr(int(v[1:], base=16))
        elif v.startswith('U+'):  # SBCL
            self.value = chr(int(v[2:], base=16))
        #elif v.startswith('M-') or v.startswith('C-'):
            # FIXME TODO XXX these require extensions to the reader
            # to distinguish cases like M-C and M-\C-A
            #self.value = v
        elif v[0] in '0123':
            pass
            # XXX racket only runs from 000 to 377 somehow ??
            # ah yes, that is 255 in octal, so only the first byte
            # then you have to use unicode
        else:
            maybe_uniname = v.replace('_', ' ')
            try:
                self.value = unilookup(maybe_uniname)
                return self.value
            except KeyError:
                pass

            raise ValueError(f'Unknown charachter name {v!r}')

        return self.value





from sxpyr._exports import *
from sxpyr.sxpyr import (
    Ast, Walk, _m,
    DispatchNotImplementedError)
from sxpyr import char
from unicodedata import name as uniname, lookup as unilookup


class WalkLazy(Walk):
    """ Bad implementations to get things working. """

    def atom(self, ast):
        v = ast.value
        if v[0].isdigit():
            try:
                # FIXME massive completely incorrect hack
                nv = literal_eval(v)  # lol so dumb
                return nv
            except Exception as e:
                pass

        return Identifier(v)


class WalkRkt(Walk):

    _recurse_funs = (
        (Syntax, '_quote'),
        (ISyntax, '_isyntax'),
    ) + Walk._recurse_funs

    _unquote_family = ('usyntax', 'susyntax') + Walk._unquote_family
    _uqf_funs = (
        (USyntax, '_uq_any'),
        (SUSyntax, '_uq_any'),
    ) + Walk._uqf_funs

    _type_funs = (
        (Syntax, 'syntax'),
        (ISyntax, 'isyntax'),
        (USyntax, 'usyntax'),
        (SUSyntax, 'susyntax'),
    ) + Walk._type_funs

    _sh_funs = (
        
    ) + Walk._sh_funs

    lists = Walk.listp
    listc = Walk.listp
    atom = WalkLazy.atom

    def __init__(self):
        super().__init__()

    def _iq_overwrite(self):
        super()._iq_overwrite()
        self.usyntax, self.susyntax = self.iq_ustx, self.iq_suns

    def _is_overwrite(self):
        self.uquote, self.suquote = self.is_uqot, self.is_sunq
        self.usyntax, self.susyntax = self.is_ustx, self.is_suns

    def _isyntax(self, ast):
        return self._quote_family_wrapper(ast, self._is_overwrite)

    def atom(self, ast):
        # NOTE since we don't have a full implementation we can't
        # define these dynamically so we hardcode them as errors here
        # this matches the behavior of calling read in racket
        if ast.value in ('#',):
            raise SyntaxError(f'Illegal atom {ast}')
        else:
            return super().atom(ast)

    def syntax   (self, ast): return ast
    def isyntax  (self, ast): return ast

    def is_ustx  (self, ast): return ast
    def is_suns  (self, ast): return ast
    def is_uqot  (self, ast): return ast
    def is_sunq  (self, ast): return ast

    def iq_ustx  (self, ast): return ast
    def iq_suns  (self, ast): return ast
    # NOTE unlike common lisp, racket only raises an error if the un
    # forms are used at the top level, if they are used inside of a
    # quoted form it will not complain not sure whether that is
    # intentional or not
    def usyntax  (self, ast):
        raise SyntaxError(f'unsyntax not in quasisyntax {ast}')
    def susyntax (self, ast):
        #breakpoint()
        raise SyntaxError(f'unsyntax splicing not in quasisyntax {ast}')

    #def sharp(self, ast):
    #    """ unspecified dispatch macros """
    #    value = ast.value
        
    def sh_atom(self, ast):
        value = ast.value
        if value == 't':
            return Boolean(True)
        elif value == 'f':
            return Boolean(False)
        elif value == 'lang':
            # FIXME ambiguity in the parser #lang at-exp racket/base
            # and other things means that the lang line has to be
            # handled directly in the reader which is extremely
            # annoying it should just slurp the next expression which
            # could be a listy thing #lang(at-exp racket/base) and
            # then it could be anywhere it would need to be clear that
            # it wasnot a function but it would definitely be more
            # consistent and easier to parse, not as aesthetic sigh
            # the balance of implementation difficulty vs additional
            # work for users vs consistency and predictability
            return LangLine('sigh-reader-is-broken-for-this')
        elif value in ('rx'):
            return Sharp(ast)  # FIXME TODO broken at the parse level
        elif value == 'rx#':  # FIXME bytestring
            # TODO in theory we could have these parse the way feature
            # expressions do so that when they terminate they don't
            # transition back to the top but instead to the OM NOM NOM
            # for the next single expression, of course the requirement
            # that there be no whitespace is artifical as far as I can
            # tell because racket always generats an error if these are
            # present alone this is true for #hash as well
            return Sharp(ast)  # FIXME TODO broken at the parse level
        elif value in ('px'):  # has different syntax perl maybe?
            return Sharp(ast)  # FIXME TODO broken at the parse level
        elif value == 'px#':  # FIXME bytestring
            return Sharp(ast)  # FIXME TODO broken at the parse level
        elif value.startswith('%'):
            # e.g. #%kernel
            return Identifier('#' + value)
        elif value.startswith('&'):
            return Box(value)
        elif value[0] in 'xdb':  # XXX warning risk of the empty atom
            return Number(ast)
        elif value == 's':
            return Struct(value)
        elif value.startswith('hash'): # FIXME FIXME
            return Dict(value)
        elif value.startswith('<<'):  # FIXME very broken
            return Sharp(ast)  # FIXME TODO broken at the parse level

        #elif value in ('',):

            # XXX incidentally the writing below is wrong, and the
            # way I wrote the reader you don't deal with a bare sharp
            # as a sharp itself, it is just an atom, we COULD implement
            # this as a Sharp ast node in caste_sharp, but it simplifies
            # things for the Sharp ast node to never be empty

            # a bare sharp is not an unbound identifier in racket it
            # is its own syntax error, that may be implemented as a
            # macro in the racket reader implementation but since we
            # don't have a complete runtime backing this we can't
            # quite pull that off yet, so we create explicit syntax
            # errors here instead, NOTE double sharp ## is not here
            # for now because the racket reader supports dispatch
            # macro chars

            #raise SyntaxError(f'Bad syntax {ast}')
        else:
            # TODO #-1 #1 #2
            # TODO arbitrary multidispatch key-5
            raise DispatchNotImplementedError(f'No dispatch transformer for {ast}')

    def sh_keyw (self, ast): return ast
    def sh_estr (self, ast):
        # TODO bytestring
        return ast
    def _sh_lst (self, ast): return Vector.from_ast(ast)
    def sh_lstp (self, ast): return self._sh_lst(ast)
    def sh_lsts (self, ast): return self._sh_lst(ast)
    def sh_lstc (self, ast): return self._sh_lst(ast)

    # FIXME need _sh_iquote in preprocessing
    def sh_quot (self, ast): return Syntax.from_ast(ast)
    def sh_iqot (self, ast): return ISyntax.from_ast(ast)
    def sh_uqot (self, ast): return USyntax.from_ast(ast)
    def sh_sunq (self, ast): return SUSyntax.from_ast(ast)

    def charspec(self, ast):
        # https://docs.racket-lang.org/reference/reader.html#%28part._parse-character%29
        value = ast.value
        if len(value) == 1:  # racket is utf8 aware by default
            # NOTE raw whitespace will not be printed the same
            # way it is in racket, but will match python repr
            nv = value
        elif value.lower() in char.known_multi:
            # racket does downcase these #\SpAcE
            nv = char.known_multi[value.lower()]
        elif value[0] == 'u' and len(value) <= 5:
            nv = chr(int(value[1:], base=16))
        elif value[0] == 'U' and len(value) <= 7:
            nv = chr(int(value[1:], base=16))
        elif value[0] in '0123' and len(value) == 3:
            # octal notation must have leading zeros in racket if it
            # is for chars, in strings it matches python
            i = int(value, base=8)
            if i > 255:
                # this matches python's behavior exactly but we have
                # to catch it here because python's chr function
                # supports arbitrary ints
                msg = ('Racket does not support octal notation '
                       'for chars longer than a single byte. '
                       'Aka 255, 0o377 or #\\377.')
                raise SyntaxError(msg)

            nv = chr(i)
        else:
            msg = f'unknown char spec #\\{value}'
            raise SyntaxError(msg)

        return Char(nv)


class WalkCl(Walk):
    def sh_lsts(self, ast):
        return Vector.from_ast(ast)


class CharEl(Ast):
    """ Emacs has some ... non-standard char representations. """

    def __init__(self, collect, elc_int=None):
        if collect and elc_int:
            raise TypeError("can't have both collect and elc_int")
        self.collect = collect
        self._elc_int = elc_int

    def __eq__(self, other):
        return (type(self) == type(other) and
                self.elc_int == other.elc_int)

    @property
    def elc_int(self):
        """ emacs char int """
        if self._elc_int is not None:
            return self._elc_int

        ints = [ord(c) if isinstance(c, str)
                else c.bit for c in self.collect]
        return sum(ints)


class WalkEl(Walk):

    @staticmethod
    def _compute_escape_value(value):
        if value in char.cee_el:
            return char.cee_el[value]

        return value

    def lists(self, ast):
        return Vector.from_ast(ast)

    def sh_lsts(self, ast):
        return ByteCode.from_ast(ast)

    def sh_atom(self, ast):
        # racket
        # https://docs.racket-lang.org/reference/reader.html#%28part._parse-number%29

        # cl
        # https://www.cs.cmu.edu/Groups/AI/html/cltl/clm/node17.html
        # see the compatibility note for further crazytown

        # el
        # https://www.gnu.org/software/emacs/manual/html_node/elisp/Integer-Basics.html

        # clj (doesn't go here, goes in regular atom)
        # https://clojure.org/guides/learn/syntax#_numeric_types

        # cl el racket scheme all seem to support this
        # hy supports python's number syntax 0b 0o 0x
        # clojure has some insane mishmash of java and sharpless cl style
        # #b -> bin literal
        # #o -> oct literal
        # #x -> hex literal

        # cl and el support #NrXXXX syntax
        return ast

    def charspec(self, ast):
        # the elisp unescaped char spec is actually pretty easy
        # everything else is hard
        value = ast.value
        if len(value) == 1:
            return Char(value)
        else:
            raise SyntaxError(f'Unknown char spec ?{value}')

    def echarspec(self, ast):
        collect = ast.collect
        c0 = collect[0]
        rest = collect[1:]
        if isinstance(c0, SEscape):
            v0 = c0.value(self._compute_escape_value)
            if isinstance(v0, type) and issubclass(v0, char.KeyChar):
                if v0 == char.SpaceOrSuperNeedDash and not rest:
                    nv = ' '  # spacceeee
                elif v0 == char.UnicodeName and rest[0] == '{' and rest[-1] == '}':
                    # this is where emacs allows variable length unicode
                    uniname = ''.join(rest[1:-1])
                    if uniname.startswith('U+'):
                        nv = chr(int(uniname[2:], base=16))
                    else:
                        try:
                            nv = unilookup(uniname)
                        except KeyError as e:
                            raise SyntaxError(f'unknown unicode name {uniname}')
                else:
                    # expect multiple escapes
                    # control-and-friends-char syntax
                    # emacs uses the 2**22 through 2**27 bits to
                    # mark that certain keys have been pressed
                    # NOTE this section is recursive, the final
                    # key of the combination can be represented
                    # however you like
                    dialect_escapes = self._compute_escape_value
                    hrm = [c.value(dialect_escapes)
                           if isinstance(c, SEscape)
                           else c for c in collect]
                    orig = collect
                    next_dash = False
                    collect = []
                    coll_str = []
                    #print('wathrm:', hrm)
                    for v, o in zip(hrm, orig):
                        #print(v)
                        # FIXME if a char came from somewhere other than
                        # the elisp reader, we could make a mistake here
                        if isinstance(v, type) and issubclass(v, char.NeedDash):
                            # FIXME still need instances here to hold points ?
                            collect.append(v)
                            next_dash = True
                        elif next_dash and v != '-':
                            raise SyntaxError(f'lol {ast}')
                        elif next_dash:
                            next_dash = False
                        elif isinstance(v, str):
                            coll_str.append(v)
                        elif v == char.UnicodeName:
                            _se = SEscape('N')
                            _se._point_beg = o._point_beg
                            _se._point_end = o._point_end
                            coll_str.append(_se)
                        elif coll_str:
                            # once you are in the final char spec
                            # you can't go back out
                            raise SyntaxError(f'oops {ast}')
                        else:
                            collect.append(v)

                    regular_char = False
                    if coll_str:  # TODO else error ??
                        if not collect:
                            regular_char = True
                        #print(coll_str)
                        if isinstance(coll_str[0], SEscape):
                            _ecs = ECharSpec(coll_str)
                            _ecs._point_beg = coll_str[0]._point_beg
                            _ecs._point_end = coll_str[0]._point_end
                            # FIXME still broken I think
                            collect.append(self.echarspec(_ecs).value)
                        else:
                            collect.append(self.charspec(CharSpec(coll_str)).value)

                    if regular_char:
                        nv = collect[0]
                    else:
                        wlk = CharEl(collect)
                        self._loc(wlk, ast)
                        return wlk

            elif isinstance(v0, str):
                vrest = ''.join(rest)  # FIXME TODO fails here -> syntax error
                # FIXME ?\^\n != ?\^n and is a valid value ;_; tears
                if v0 == 'x' and len(vrest) <= 7:
                    # ?\xfffffff is max and may have variable width
                    # ?\x seems to read as ?\x0 ?? shouldn't that signal an error ??
                    # NOTE python tops out at 0x110000 which is 1114111 which is the
                    # current max unicode char emacs just goes wild apparently
                    i = int(vrest, base=16)
                    try:
                        nv = chr(i)
                    except ValueError:
                        wlk = CharEl(None, elc_int=i)
                        self._loc(wlk, ast)
                        return wlk

                if v0 == 'u' and len(vrest) == 4:
                    # ?\uffff is max and fixed width
                    nv = chr(int(vrest, base=16))
                elif v0 == 'U' and len(vrest) == 8:
                    # elisp U must be 8 hex
                    # ?\U0010ffff seems to be max for the extended unicode
                    # but is not actually a valid unicode char right now
                    # python matches the error behavior here
                    nv = chr(int(vrest, base=16))
                else:
                    value = v0 + vrest
                    if not vrest:
                        nv = value
                    elif v0 in '01234567' and len(value) <= 3:
                        # emacs supports up to 777 aka 511 aka 512 or 9 bits
                        # but unlike racket does not require leading zeros
                        i = int(value, base=8)
                        if i > 511:
                            msg = ('Elisp does not support octal notation '
                                   'for chars longer than 9 bits. '
                                   'Aka 511, 0o777 or ?\\777.')
                            raise SyntaxError(msg)

                        nv = chr(i)

            else:
                raise SyntaxError(f'unknown char spec {ast}')
        else:
            # someone is up to some weirdness and this probably
            # isn't a real char spec
            raise SyntaxError(f'wat char spec {ast}')
            
        wlk = Char(nv)
        self._loc(wlk, ast)
        return wlk

        #return ast.value(lambda v:v, self._compute_escape_value)


class WalkUnion(Walk):
    """ All together now! """



from .sxpyr import (
    configure,
    make_do_path,
    __version__)

# ast nodes
from .sxpyr import (
    ListP,
    ListS,
    ListC,
    Keyword,
    EKeyword,
    Atom,
    EAtom,
    Escape,
    SEscape,
    String,
    EString,
    CharSpec,
    ECharSpec,
    Comment,
    BComment,
    XComment,
    HComment,
    Quote,
    IQuote,
    UQuote,
    SUQuote,
    Syntax,
    ISyntax,
    USyntax,
    SUSyntax,
    Sharp,
    FeatureExpr,)

# weird stuff
from .sxpyr import (
    Identifier,
    LangLine,)

# data types
from .sxpyr import (
    Cons,
    List,
    Array,
    Vector,
    Set,
    Dict,
    Char,
    Boolean,)

# dialect configs
from .sxpyr import (
    conf_plist,
    conf_sxpyr,
    conf_union,
    conf_cl,
    conf_el,
    conf_xel,
    conf_rkt,
    conf_gui,
    conf_clj,
    conf_hy,
    conf_txr,)



import re
from setuptools import setup


def find_version(filename):
    _version_re = re.compile(r"__version__ = ['\"](.*)['\"]")
    last = None  # match python semantics
    for line in open(filename):
        version_match = _version_re.match(line)
        if version_match:
            last = version_match.group(1)

    return last


__version__ = find_version('sxpyr/sxpyr.py')

with open('README.org', 'rt') as f:
    long_description = f.read()

cli_requires = ['pyontutils']  # FIXME should be orthauth but haven't moved clifun yet
fuzz_requires = ['python-afl'] + cli_requires
tests_require = ['pytest'] + cli_requires
                 
setup(name='sxpyr',
      version=__version__,
      description='A flexible Lisp reader.',
      long_description=long_description,
      long_description_content_type='text/markdown',  # FIXME org -> markdown
      url='https://github.com/tgbugs/sxpyr',
      author='Tom Gillespie',
      author_email='tgbugs@gmail.com',
      license='MIT',
      classifiers=[
          'Development Status :: 3 - Alpha',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
          'Programming Language :: Python :: 3.10',
          'Programming Language :: Python :: 3.11',
          'Programming Language :: Python :: Implementation :: CPython',
          'Programming Language :: Python :: Implementation :: PyPy',
          'Operating System :: POSIX :: Linux',
          'Operating System :: MacOS :: MacOS X',
          'Operating System :: Microsoft :: Windows',
      ],
      keywords=('lisp reader edn clojure racket scheme '
                'elisp emacs sexp s-expression parser parsing'),
      packages=[
          'sxpyr',
      ],
      python_requires='>=3.7',
      tests_require=tests_require,
      extras_require={'test': tests_require,
                      'cli': cli_requires,
                      'fuzz': fuzz_requires,
                     },
      scripts=[],
     )



# parse
# walk
# caste

# TODO attach source char and line to the syntax objects
# TODO issues with ambiguity of #word(list 1 2 3)
# TODO feature expressions aka wraps next n, the impl must be
# chainable and local because n must be defined, practically this ends
# up being a wraps next that transitions the state to a wraps next

__version__ = '0.0.2'

from io import TextIOWrapper
from ast import literal_eval
from json import dumps
from types import GeneratorType

debug = False


class SxpyrError(Exception): pass
class UnknownStateError(SxpyrError): pass
class DispatchNotImplementedError(SxpyrError): pass


def python_to_sxpr(blob):
    # maybe json to sxpr to be safe? use the json normalization that we have?
    # blob to sxpr?

    # racket has jsexprs but that is peculiar to racket's syntax
    # and makes uses of verbatim atoms for keys so that it is possible
    # to have leading colons in keys etc. we can't do that with plists
    # and be able to use these in common lisp due to the ::symbol issue
    if isinstance(blob, dict):
        return PList([e for k, v in blob.items() for e in [Keyword(k)._set_bounds(), python_to_sxpr(v)]])
    elif isinstance(blob, list):
        return List([python_to_sxpr(v) for v in blob])
    elif blob is None:
        return List([])
    else:
        return blob


def print_plist(ast_dt):
    if isinstance(ast_dt, LLike):
        if ast_dt.value:
            #if Keyword in [type(v) for v in ast_dt.value]:
                #return '(\n', ')\n'
            #else:
            return '(', ')\n'
        else:
            return '()'
    elif isinstance(ast_dt, Keyword):
        return '\n:'
    elif isinstance(ast_dt, str):
        bad = '()[]{} \n\t\'\"'
        if [c for c in bad if c in ast_dt]:
            return dumps(ast_dt)
        else:
            # make it a symbol
            return ast_dt
    elif isinstance(ast_dt, int):
        return str(ast_dt)
    else:
        breakpoint()
        raise NotImplementedError(type(ast_dt))


def make_do_path(do, chunksize=4096):
    """ along the way to load """
    #parse = configure(**kwargs)  # FIXME pass in parse
    #read = conf_read(parse, mawp)  # FIXME
    def do_path(path_or_fd):
        if isinstance(path_or_fd, TextIOWrapper):  # stdin probably
            def path_gen():
                f = path_or_fd
                while True:
                    data = f.read(chunksize)
                    if not data:
                        break
                    yield from data
        else:
            def path_gen():
                path = path_or_fd
                with open(path, 'rt') as f:
                    while True:
                        data = f.read(chunksize)
                        if not data:
                            break
                        yield from data

        return do(path_gen())

    return do_path


def conf_read(parser, walk_cls):
    #def skip(v):
        #return [c for c in (Comment, XComment, BComment)
                #if isinstance(v, c)]

    def _walk(ast):
        #print('walk:', ast)
        if [_ for _ in (Sharp, Quote, IQuote, UQuote, SUQuote)
            if isinstance(ast, _)]:
            revalue = walk_help(ast.value, walk)
            ast.value = revalue
        elif isinstance(ast, ListAbstract):
            # FIXME this kind of defeats the point of having
            # that nice pda in the first place when we recursively
            # descend here
            recollect = [walk_help(_, walk) for _ in ast.collect
                         if not skip(_)]
            ast.collect = recollect  # FIXME not quite right

        return walk_help(ast, None)
        # FIXME TODO quote and such
        #if ast.__class__ in mawp:
            #callable_thing = mawp[ast.__class__]
            #return callable_thing(ast)
        #else:
            #return ast

    def read(gen):
        walk = walk_cls()  # avoid any whiff of threading issues
        exp_gen = parser(gen)
        for expression in exp_gen:
            if debug:
                print('read:', expression)
            if not walk._skip(expression):
                transformed = walk(expression)
                yield transformed

    return read


class _m:
    """ helper methods"""

    def eq_value(self, other):
        return type(self) == type(other) and self.value == other.value

    def eq__value(self, other):
        return type(self) == type(other) and self._value == other._value

    def eq_collect(self, other):
        # FIXME interned vs uninterned
        return type(self) == type(other) and self.collect == other.collect


# abstract syntax tree node types

class Ast:

    def _set_bounds(self, beg=None, end=None):
        self._point_beg = beg
        self._point_end = end
        return self

    def __repr__(self, **kwargs):
        if not hasattr(self, '_point_beg'):
            # FIXME _point_beg and _point_end may not be know if the object
            # is being printed instead of read ...
            breakpoint()
        pb, pe = self._point_beg, self._point_end
        pts = f' ::{pb}:{pe}' if debug else ''
        if hasattr(self, 'value') and type(self.value) != type(self.__repr__):
            if isinstance(self.value, String):
                return f'<{self.__class__.__name__[:2]} {self.value!r}{pts}>'
            else:
                return f'<{self.__class__.__name__[:2]} {self.value}{pts}>'
        if hasattr(self, '_value'):
            return f'<{self.__class__.__name__[:2]} {self._value}{pts}>'
        elif hasattr(self, 'collect'):
            return f'<{self.__class__.__name__[:3]} {self.collect!r}{pts}>'
        else:
            #self.__class__.__repr__ = lambda s: 'debug'
            #breakpoint()
            raise Exception('oops')

    def caste(self, typef):
        """ Recursively caste all nested forms.

        the final form of the ast must have been
        achieved so that it has a defined value """
        # if your ast is not fully resolved this will cause an error
        # down the line, so we just error here

        if type(self.value) == type(self.__repr__):
            raise TypeError(
                'self.value has not been fully factored yet! '
                'This usually happens if self.value is a function that takes.'
            )

        return typef(self.value)

    def __hash__(self):
        return hash((self.__class__, self.value))

        # Atom

        # FIXME need clarity on what identity model we follow here wrt
        # where Atom is defined due to python issues with __main__ vs
        # module imports and lisp interned vs uninterned

        # XXX NOTE this only compares the Ast values !!!!
        # it does not deal with the fact that there needs
        # to be a second pass where atoms are processed
        # further into numbers, symbols, etc. we might even
        # consider moving keywords to this phase as well
        # XXX NOTE this also means that symbol names have
        # NOT BEEN RESOLVED AT THIS POINT!

        # Keyword

        # FIXME need clarity on what identity model we follow here wrt
        # where Keyword is defined due to python issues with __main__
        # vs module imports keywords are self interning or somthing
        # like that so they are not as big an issue as symbols/atoms

    @classmethod
    def from_ast(cls, ast):
        # FIXME TODO homogenous would be to pass collect
        # along and/or track the ast that it came from
        # so that the loc data can travel along
        return cls(ast.value)


class ListAbstract(Ast):

    __eq__ = _m.eq_collect

    @classmethod
    def from_elements(cls, *elements):
        return cls(list(elements))

    def __init__(self, collect):
        self.collect = collect

    def __repr__(self):
        return f'<{self._o} {repr(self.collect)[1:-1]} {self._c}>'

    @property
    def value(self):
        # FIXME !??!
        return self


class ListP(ListAbstract):
    _o, _c = '()'


class ListS(ListAbstract):
    _o, _c = '[]'


class ListC(ListAbstract):
    _o, _c = '{}'


class Keyword(Ast):

    # swindle uses
    # #: -> real-keyword
    # : -> keyword
    # ' -> symbol

    # FIXME pretty sure we should get rid of these in the read pass
    # and let the individual implementations determine? but trickly
    # the tradeoff is that we have to make sure that we can
    # distinguish :|a b| and |:a b| internally if we read both
    # of them into Atom

    __eq__ = _m.eq_value
    __hash__ = Ast.__hash__  # would someone care to explain to me why
                             # this doesn't inherit ???
                             # https://bugs.python.org/issue1549 sigh
                             # so much for trying to be performant
                             # just use branches inside the methods
                             # to reduce the book keeping! SIGH

    @classmethod
    def from_collect(cls, collect):
        self = cls(''.join(collect))
        self.collect = collect
        return self

    def __init__(self, value):
        self.value = value

    def _print(self, pf):
        beg = pf(self)
        return beg + self.value


class EKeyword(Keyword):

    __eq__ = _m.eq_collect

    def __init__(self, collect):
        self.collect = collect

    def value(self, dialect_ak_escape):
        if dialect_ak_escape is None:
            dialect_ak_escape = dialect_ak_escape_def

        return dialect_ak_escape(self.collect)


class Atom(Ast):
    """ Numbers, identifiers, booleans, and more.
        These are left uninterpreted. """

    __eq__ = _m.eq_value

    def __init__(self, collect):
        self._collect = collect
        self.value = ''.join(collect)
        # TODO convert the atoms to python types
        '#t' '#f' 't' 'nil' '1234'


def dialect_ak_escape_def(c):
    return ''.join([_.value
                    if isinstance(_, Escape)
                    else _ for _ in c])


class EAtom(Atom):
    # TODO FIXME figure out whether EThings are really subClassOf Things
    # or whether that will cause issues, in a sense, yes they are Chars
    # but they have different behavior and requirements

    __eq__ = _m.eq_collect

    def __init__(self, collect):
        self.collect = collect

    def value(self, dialect_ak_escape=None):
        if dialect_ak_escape is None:
            # these nearly always go to identifiers I think?
            dialect_ak_escape = dialect_ak_escape_def

        return dialect_ak_escape(self.collect)


class EscAbstract(Ast):

    __eq__ = _m.eq__value

    def __init__(self, value):
        self._value = value

    def __hash__(self):
        return hash((self.__class__, self._value))


class Escape(EscAbstract):
    """ Symbol escapes """

    @property
    def value(self):
        # XXX NOTE THE DIFFERENCE FROM SEscape
        # symbol escapes do not have special
        # meaning, even in elisp, they just
        # prevent any special behavior from occuring
        # e.g. the end of a symbol or the start of a
        # number
        return self._value


class SEscape(EscAbstract):
    """ String and char escape. """

    def value(self, dialect_escapes):
        return dialect_escapes(self._value)


class String(str, Ast):
    """ Ast for plain strings so that we can
        do things like track start/end """

    @property
    def value(self):
        return str(self)

    def __repr__(self):
        return dumps(self)


class EString(Ast):
    """ This is an unprocessed string that is still
    made out of its various parts. If you encounter
    an ast node of this type it should contain escaped
    charachters that need to be processed. """

    __eq__ = _m.eq_collect

    def __init__(self, collect):
        self.collect = collect

    def value(self, dialect_escapes):
        ''.join([c.value(dialect_escapes)
                 if isinstance(c, SEscape)
                 else c for c in self.collect])


class CharSpec(Ast):

    __eq__ = _m.eq_value

    def __init__(self, collect):
        self.value = ''.join(collect)

    #def value(self, dialect_char):
        #return dialect_char(self._value)


class ECharSpec(CharSpec):
    """ chars with escape sequences """

    __eq__ = _m.eq_collect

    def __init__(self, collect):
        self.collect = collect

    def value(self, dialect_char, dialect_escapes):
        value = ''.join([c.value(dialect_escapes)
                         if isinstance(c, SEscape)
                         else c for c in self.collect])

        # FIXME a bit more complex than this I think
        # maybe need dialect_char_esc or similar?
        return dialect_char(value)


class Comment(Ast):
    """ Line comment """

    __eq__ = _m.eq_value

    def __init__(self, collect):
        self.collect = collect
        self.value = ''.join(collect)


class BComment(Ast):
    """ Block comment """

    __eq__ = _m.eq_value

    def __init__(self, collect):
        self.collect = collect
        self.value = ''.join([v._nest if isinstance(v, self.__class__) else v for v in collect])

    @property
    def _nest(self):
        return f'#|{self.value}|#'


class WrapsNext(Ast):

    __eq__ = _m.eq_value

    @classmethod
    def from_collect(cls, collect):
        comments = [c for c in collect if isinstance(c, Comment) or isinstance(c, XComment)]
        collect = [c for c in collect if not (isinstance(c, Comment) or isinstance(c, XComment))]
        if not collect:
            # FIXME not all wraps next things require content to follow ?
            raise SyntaxError(f'Empty {cls.__name__}!')  # FIXME context for debug
        assert len(collect) == 1, collect
        self = cls(collect[0])
        if comments:
            self.comments = comments

        return self

    def __init__(self, value):
        self.value = value


class Quote(WrapsNext):
    """ HI there. """


class IQuote(WrapsNext):
    """ Quasiquote """


class UQuote(WrapsNext):
    """ Unquote """


class SUQuote(WrapsNext):
    """ Splicing Unquote """


class Sharp(WrapsNext):
    """ Sharp! Hash! Octothorp! Who knows. """
    # FIXME rename to Dispatch


class XComment(WrapsNext):
    """ eXpression comments, aka datum comments """


class HComment(WrapsNext):
    """ Header comments, axiomatically at the top level of a file. """


class WrapsNextN(WrapsNext):
    """ wrap the next n expressions """
    # these actually have to be done in the first pass right now
    # because they eat the next fixed n expressions and grouping
    # otherwise independent expressions is only as part of the first
    # phase of the reader, that grouping could be done in the second
    # phase as well, but that just increases the complexity of both
    # phases needlessly note that this has to be implemented using
    # only local transition rules and thus the explicit n is implicit
    # in the structure of the possible state transition in the parser


class FeatureExpr(WrapsNextN):
    """ Common Lisp feature expressions """


# one and a halfth phase? or 2nd phase? these are things that could be
# parsed directly from the stream, but are not because they are
# lurking behding the sharp and we don't interpret the sharp until
# phase two in most dialects, this doesn't have to be the case if
# there is the single token transition, more correctly, these would be
# named DispatchQuote, DispatchQuasiQuote, Dispatch


class Syntax(WrapsNext):
    """ In principle one could have a single token that transitioned to syntax. """
    __eq__ = _m.eq_value
    def __init__(self, value): self.value = value


class ISyntax(WrapsNext):
    __eq__ = _m.eq_value
    def __init__(self, value): self.value = value


class USyntax(WrapsNext):
    __eq__ = _m.eq_value
    def __init__(self, value): self.value = value


class SUSyntax(WrapsNext):
    __eq__ = _m.eq_value
    def __init__(self, value): self.value = value


# second phase abstract types (incomplete)
# not clear we need this, I think it is probably
# one layer of indirection too many


class PhaseTwoThing:
    """ Not really an ast anymore. """


class AtomsChildren(PhaseTwoThing):
    """ Any number of things here """


class Identifier(AtomsChildren):
    __eq__ = _m.eq_value
    def __init__(self, value): self.value = value


class Boolean(AtomsChildren):
    __eq__ = _m.eq_value
    def __init__(self, value): self.value = value


class Number(AtomsChildren):
    __eq__ = _m.eq_value
    def __init__(self, value): self.value = value


class LangLine(PhaseTwoThing):
    """ the racket lang line #lang """
    # XXX This is comment-like or WrapsUnti this and HereDoc could
    # work by shifting the state to be sensitive until the next
    # newline, but otherwise the same, the problem is that you can't
    # eastily use a phased approach because you won't ever be able
    # to turn off the process of reading whole lines, a bit clearer
    # how one might push arbitrary values to the PDA stack to do this
    __eq__ = _m.eq_value
    def __init__(self, value): self.value = value


# datatypes, mostly concrete


class DataType:  # FIXME naming
    """ objects that have specific semantics in certain dialects
    this provides a layer of indirection so that the choice of
    the python structure used internally can be tailored to the
    exact use case without blindly following upstream """


class Char(DataType):
    """ the char spec fully processed down
    to a single charachter or value """
    __eq__ = _m.eq_value
    def __init__(self, value): self.value = value


class Cons(DataType):
    """ A single cell in a singly linked list, usually mutable. """


class LLike(DataType):

    o, c = None, None
    __eq__ = _m.eq_value

    @classmethod
    def from_ast(cls, ast):
        # FIXME TODO homogenous would be to pass collect
        # along and/or track the ast that it came from
        # so that the loc data can travel along
        return cls(ast.collect)

    def __init__(self, value):
        self._value = value

    @property
    def value(self):
        return self._value

    def caste(self, typef):
        """ caste should return """
        value = [ast.caste(typef) if isinstance(ast, Ast)
                 else ast for ast in self.value]
        # descend first the caste
        return typef(self.__class__(value))

    def __repr__(self, depth=0, **kwargs):
        # already have a printer for this in protcur
        # FIXME recursion errors for deeply nested structures
        r = '\n'.join(repr(_) for _ in self.value)
        return self.o + r + self.c

    def _print(self, pf):
        beg, end = pf(self)
        return beg + ' '.join([c._print(pf)
                               # FIXME sigh type inhomogenaity in the IR ...
                               if hasattr(c, '_print') else
                               pf(c)
                               for c in self.value]) + end


class List(LLike):
    """ Usually a mutable singly linked list. """

    o, c = '()'


class Array(LLike):
    """ An n-dimensional array. """

    o, c = '(array ', ')'


class Vector(LLike):
    """ A one dimensional array of fixed length.
    Access is expected to be O(1). """

    o, c = '[]'


class Set(LLike):
    """ A mathematical set. """

    o, c = '#{', '}'


class PList(LLike):
    """ In cl it is (getf a :b) and (setf (getf a :b) 'c)."""
        # In el use getf/setf to act directly as well. Oddly plist-get
        # and plist-put do not do what you expect? Ah. Because unlike
        # setf/getf they expect the value to already be present? That
        # is NOT what the docs say. Ah but the plist cannot be empty
        # to start with. Very strange. CL doesn't do this.

    o, c = '(plist (', '))'

    def __init__(self, value):
        if len(value) % 2 != 0:
            breakpoint()
            # FIXME need the point location report for this
            raise ValueError('plist is not a plist!')

        self._value = value


class Dict(LLike):
    """ Sometimes called a hash, sometimes called a map.
    Expectation is that access is O(1). """

    o, c = '{', '}'


class ByteCode(LLike):
    """ elisp can read bytecode literals """

    o, c = '#[', ']'


class Struct(LLike):
    """ literal for abstract data type container thing """
    # the old slots vs keys thing, but we're over that

    o, c = '(struct', ')'   # XXX TODO

class Box(DataType):
    """ indirection wrapper """

    o, c = '(box', ')'   # XXX TODO

    __eq__ = _m.eq_value
    def __init__(self, value): self.value = value


def plist_to_dict(plist):
    # use with Ast.caste XXX NOTE Ast.caste takes care of the
    # recursion for us this is slightly inefficient because we have to
    # walk over everything twice, but that is expected given the way
    # this is broken into three phases for clarity
    if len(plist) % 2 != 0:
        raise SyntaxError('plist is not a plist')
    try:
        return dict(zip(plist[::2], plist[1::2]))

        # since python only has strings and this is not intended to be
        # written back out ... we aren't going to worry about the fact
        # that {:key value} and {"key" value} are conflated because
        # meshing with the python ergonimics is more important will
        # have to add a note about this, hy is in the same situation
        # where it has to keep them wrapped in the ast, same issue for
        # atoms on the other side, but that is a different issue
        # unfortunately to do this right it has to be done in the
        # caste step
        #return {a.value:b for a, b in zip(plist[::2], plist[1::2])}
    except TypeError as e:
        breakpoint()
        pass


# class for pda states

class State:

    state_names = {}  # update this with all the dialect variants of the names

    def __init__(self, index):
        self.index = index

    def __repr__(self):
        try:
            return f'<state {self.state_names[self]}>'
        except KeyError as e:
            breakpoint()
            pass


# the unparsable char

unp = object()  # unparsable  # cannot use None

# helper for parsing elisp chars

escape_hat = SEscape('^')  # needed to parse elisp escape for control ?\^?

# union tokens

class UnionToken:
    def __init__(self, *chars):
        self.chars = chars

    def __eq__(self, other):
        return other in self.chars


union_t_to_unquote = UnionToken(',', '~')
union_t_to_xcom_in_shrp = UnionToken('_', ';')
union_t_to_feat_in_shrp = UnionToken('+', '-')

# tokens

toks_common = {
    't_newline':             '\n',
    't_tab':                 '\t',
    't_space':               ' ',
    't_beg_list_p':          '(',
    't_end_list_p':          ')',
   #'t_beg_list_s':          '[',
   #'t_end_list_s':          ']',
   #'t_beg_list_c':          '{',
   #'t_end_list_c':          '}',
    't_beg_end_str':         '"',
   #'t_beg_end_aver':        '|',
    't_to_sharp':            '#',
   #'t_to_char_in_shrp':     '\\',  # TODO
   #'t_to_cblk_in_shrp':     '|',
   #'t_to_xcom_in_shrp':     ';',
   #'t_to_hcom_in_shrp':     '*',
   #'t_to_keyw_in_shrp':     ':',
   #'t_to_feat_in_shrp':     union_t_to_feat_in_shrp,
   #'t_to_strn_in_shrp':     '[',  # sigh hy TODO
    't_to_quote':            "'",
    't_to_quasi':            '`',
    't_to_unquote':          ',',
   #'t_to_unquote_in_quasi': ',',  # FIXME get this sorted out vs unquote anywhere
    't_to_splc_in_unq':      '@',
   #'t_to_keyw':             ':',  # FIXME racket and scheme keywords don't do this
   #'t_to_esc':              '\\',
    't_to_esc_in_str':       '\\',
   #'t_to_char':             '?',
    't_to_comment':          ';',
}

toks_b_list_s = {
    't_beg_list_s': '[',
    't_end_list_s': ']',
}

toks_b_list_c = {
    't_beg_list_c': '{',
    't_end_list_c': '}',
}

toks_b_list_sc = {
    **toks_b_list_s,
    **toks_b_list_c,
}

toks_scheme = {
    **toks_common,
    **toks_b_list_s,
    't_to_esc':          '\\',
    't_to_xcom_in_shrp': ';',
    't_beg_end_aver':    '|',
    't_to_cblk_in_shrp': '|',
    't_to_keyw_in_shrp': ':',
}

# dialect configuration

conf_plist = {
    # FIXME this will be overly accepting whereas it needs to actively
    # raise errors when it encounters a form that has multiple
    # interpretations, in that sense the minimal reader should be
    # constructed not from the intersection aka common config, but
    # instead from the union config for all tokens and error if a
    # state is entered, this means that we _do_ need a variant of the
    # parser that accepts ~ and , and #; and #_ at the same time
    **toks_common,
    't_to_keyw':         ':',  # need this for correct symbol-name behavior
}

conf_sxpyr = {
    'additional_whitespace': ',',
    **toks_common,
    **toks_b_list_sc,
    't_to_keyw':             ':',
    't_to_xcom_in_shrp':     ';',  # more consistent that #_
    't_to_unquote':          '~',  # comma as whitespace is too useful
   #'t_to_unquote_in_quasi': union_t_to_unquote,  # as nice as it would be, context sensativity is EEK
    't_to_hcom_in_shrp':     '*',  # taking a note from org-mode
    # no block quotes no escapes and no verbatim atoms
}

conf_cl = {
    **toks_common,
    't_to_keyw':         ':',
    't_to_esc':          '\\',
    't_beg_end_aver':    '|',
    't_to_cblk_in_shrp': '|',
    't_to_feat_in_shrp': union_t_to_feat_in_shrp,
    # NOTE: do not use to_keyw_in_shrp for uninterned
    # it is simple to use (sharp (keyword ...)) than
    # to try to reconfigure this part of the reader
}

conf_el = {
    **toks_common,
    **toks_b_list_s,
    't_to_keyw': ':',
    't_to_char': '?',
    't_to_esc':  '\\',
}

conf_xel = {
    # FIXME HAHA those cheeky xemacs implementers did it!
    # they made unquote only take effect inside a quasiquote!
    # everyone else made it an error, but they just make it work!
    #'allow_in_symbol': conf_el['t_to_unquote'],  # we may need to restore this
    **conf_el,
    't_to_unquote': unp,
    't_to_unquote_in_quasi': union_t_to_unquote,
}

conf_hy = {
    'eval_collection_literals': True,  # note this is really eval lists, sets, dicts
    **toks_common,
    **toks_b_list_sc,
    't_to_keyw':         ':',
    't_to_xcom_in_shrp': '_',
    't_to_unquote':      '~',
}

conf_clj = {
    'allow_in_symbol': toks_common['t_to_quote'],
    'eval_collection_literals': True,  # this is beyond the reader, cl, el, rkt, scm, are False for this
    # TODO this could also be called implicitly_quote_collections, very annoying behavior ...
    # since it breaks equivalence between #(1 2 (+ 1 2)) and (vector 1 2 (+ 1 2))
    # and implies that #() is not simply a macro that expands to (vector )
    # see https://nullprogram.com/blog/2012/07/17/
    # on the other hand, it is safer because you know that the printed output
    # of certain structures won't suddenly be able to execute arbitrary code
    # if it is evaled again, then again, if you eval the output of quote you
    # can pwn yourself much more easily, so I don't see what the issue is
    # macros that work like string vs regular macros that expand and eval
    'additional_whitespace': ',',
    **toks_common,
    **toks_b_list_sc,
    't_to_keyw':         ':',
    't_to_char':         '\\',
    't_to_xcom_in_shrp': '_',
    't_to_unquote':      '~',
}

conf_rkt = {
    **toks_scheme,
    **toks_b_list_sc,
}

conf_gui = {
    'allow_in_symbol': toks_scheme['t_to_quote'],
    **toks_scheme,
}

conf_txr = {
    **toks_common,
    **toks_b_list_s,
    't_to_quasi':       '^',
    't_to_splc_in_unq': '*',
    't_to_xcom_in_shrp': ';',
    #'t_to_binr_in_shrp': 'b',  # binary number repr
}

# union of all or as close as we can get

conf_union = {
    'additional_whitespace': ',',  # will break some xel cases
    **conf_rkt,
    't_to_keyw':             ':',
    't_to_xcom_in_shrp':     union_t_to_xcom_in_shrp,
    't_to_unquote_in_quasi': union_t_to_unquote,
}

# walk


class Walk:

    _recurse_funs = (
        (ListAbstract, '_labst'),
        (Sharp, '_sharp'),  # process this up here FIXME Dispatch has to be implemented as a macro function
        (Quote, '_quote'),  # racket allows unquote in quote
        (IQuote, '_iquote'),
        (WrapsNext, '_wraps_next'),  # lol I actually had it implemented already
        )

    _type_funs = (
        (ListP, 'listp'),
        (ListS, 'lists'),
        (ListC, 'listc'),

        (EAtom, 'eatom'),  # FIXME escapse first because they are sco the
        (Atom, 'atom'),    # non-escaped version, which is probably wrong
        (EKeyword, 'ekeyword'),
        (Keyword, 'keyword'),
        (ECharSpec, 'echarspec'),
        (CharSpec, 'charspec'),
        (EString, 'estring'),
        (String, 'string'),

        (Sharp, 'sharp'),  # for completeness, but this is rarely used
        (Quote, 'quote'),
        (IQuote, 'iquote'),
        (UQuote, 'uquote'),
        (SUQuote, 'suquote'),
    )

    _unquote_family = 'uquote', 'suquote'
    _uqf_funs = (  # quote like functions, update if there are others
        (UQuote, '_uq_any'),
        (SUQuote, '_uq_any'),
    )

    _sh_funs = (
        (ListP, 'sh_lstp'),
        (ListS, 'sh_lsts'),
        (ListC, 'sh_lstc'),

        (EAtom, 'sh_eatm'),
        (Atom, 'sh_atom'),
        (EKeyword, 'sh_ekey'),
        (Keyword, 'sh_keyw'),
        (EString, 'sh_estr'),
        (String, 'sh_strn'),

        (Quote, 'sh_quot'),
        (IQuote, 'sh_iqot'),
        (UQuote, 'sh_uqot'),
        (SUQuote, 'sh_sunq'),
    )

    def __init__(self, casters=None):
        prev_rf = tuple(self._recurse_funs)
        restore = [(fname, getattr(self, fname)) for fname in self._unquote_family]
        self._restore_stack = [(prev_rf, restore)]
        #self._uquote, self._suquote = self.uquote, self.suquote

    def __call__(self, ast, recurse=True):

        wlk = ast
        if recurse:
            for cls, attr in self._recurse_funs:
                if isinstance(ast, cls):
                    # have to use getattr so that we can dynamically
                    # reassing the function names for nested values
                    fun = getattr(self, attr)
                    wlk = fun(ast)
                    if wlk != ast:
                        self._loc(wlk, ast)
                    break  # duh

        for cls, attr in self._type_funs:
            if isinstance(ast, cls):
                fun = getattr(self, attr)
                wlk = fun(ast)
                if wlk is None:
                    breakpoint()
                elif wlk != ast:
                    self._loc(wlk, ast)
                return wlk

        # TODO do we caste here or do we walk again?
        # yes we could do this all in a single pass
        # inside the parser and not have to walk over
        # this 3 times, but it is much easier to understand
        # what is going on when we do it this way and if
        # we want it to go faster we can figure out
        # how to write the transform using this saner
        # version as a reference
        return wlk

    @staticmethod
    def _skip(ast):
        return [c for c in (Comment, XComment, BComment)
                if isinstance(ast, c)]

    @staticmethod
    def _loc(wlk, ast):
        if not hasattr(ast, '_point_beg'):
            breakpoint()
        wlk._point_beg = ast._point_beg
        wlk._point_end = ast._point_end
        return wlk

    def _labst(self, ast):
        recollect = [self(_) for _ in ast.collect
                     if not self._skip(_)]
        return self._loc(ast.__class__(recollect), ast)

    def _wraps_next(walk, ast):  # just to confuse you, walk is indeed self
        if debug:
            print('wn:', ast)
        revalue = walk(ast.value)
        return walk._loc(ast.__class__(revalue), ast)

    def _sharp(self, ast):
        # FIXME TODO do we run the first pass descent here or not? 
        # really not sure ...
        # we can't run it after because people could have changed
        # the types, if we don't run it or try to run it after or
        # change the behavior to return None to signal no change
        # then we could run the standard transformations
        # answer: two examples make it clear that we cannot
        # run these before, #() will be caste to second layer
        # incorrectly, and other issues, the right thing to do
        # is have the default be to run the unshed function so
        # that it is accessible via super if needed, by default
        # this means that the sharp is simply ignored as if it
        # does not exist
        # ast = self._wraps_next(ast)
        if debug:
            print('rsharp:', ast)
        walk = self
        value = ast.value
        for cls, attr in self._sh_funs:
            if isinstance(value, cls):
                fun = getattr(self, attr)
                ast = fun(value)
                self._loc(ast, value)
                if isinstance(ast, Sharp):  # prevent infinite recursion
                    return ast
                else:
                    # walk is applied after sharp acts locally if the
                    # the function did not simply restore the sharp
                    # because it doesn't know what to do with it at
                    # this phase
                    return walk(ast)
        else:
            raise NotImplementedError(f'missed something {ast}')

        return walk(ast)

    def _quote(self, ast):
        # fake dynamic variables, non-threadsafe 
        #prev_rf, prev_uq, prev_suq = tuple(self._recurse_funs), self.uquote, self.suquote
        prev_rf = tuple(self._recurse_funs)
        restore = [(fname, getattr(self, fname)) for fname in self._unquote_family]
        try:
            # can't += because wraps next will match before these
            self._restore_stack.append((prev_rf, restore))
            self._recurse_funs = self._uqf_funs + prev_rf
            #self.uquote, self.suquote = self.q_uqot, self.q_sunq
            # FIXME need to match function to type
            #if 'susyntax' in self._unquote_family:# and isinstance(ast, ISyntax):
                #breakpoint()
            [setattr(self, fname, self._unquote_family_in_quote)
             for fname in self._unquote_family]
            if debug: print('enter quot/stx')
            return self._wraps_next(ast)
        finally:
            self._recurse_funs = prev_rf
            [setattr(self, *fnf) for fnf in restore]
            self._restore_stack.pop()
            if debug: print('leave quot/stx')

    def _unquote_family_in_quote(self, ast): return ast

    def _quote_family_wrapper(self, ast, overwrite):
        prev_rf, prev_uqf = tuple(self._recurse_funs), tuple(getattr(self, a) for a in self._unquote_family)
        try:
            self._recurse_funs = self._uqf_funs + self._recurse_funs
            overwrite()
            #self._recurse_funs = self._q_funs + self._recurse_funs
            #self.uquote, self.suquote = self.iq_uqot, self.iq_sunq
            if debug: print('enter iqot/stx')
            return self._wraps_next(ast)
        finally:
            self._recurse_funs = prev_rf
            for fun, attr in zip(prev_uqf, self._unquote_family):
                setattr(self, attr, fun)
            if debug: print('leave iqot/stx')

    def _iq_overwrite(self):
        #self._recurse_funs = self._q_funs + self._recurse_funs
        self.uquote, self.suquote = self.iq_uqot, self.iq_sunq

    def _iquote(self, ast):
        return self._quote_family_wrapper(ast, self._iq_overwrite)

    def _uq_any(self, ast):
        # fake dynamic variables, non-threadsafe 
        # FIXME shouldn't apply this the quote case?
        # FIXME needs stack access I think?
        #prev_rf, prev_uq, prev_suq = tuple(self._recurse_funs), self.uquote, self.suquote
        c_prev_rf = tuple(self._recurse_funs)
        c_restore = [(fname, getattr(self, fname)) for fname in self._unquote_family]
        prev_rf, restore = self._restore_stack[-1]  # FIXME still not quite right I think?
        try:
            self._restore_stack.append((c_prev_rf, c_restore))
            self._recurse_funs = prev_rf
            [setattr(self, *fnf) for fnf in restore]
            if debug: print('enter uq/sot')
            return self._wraps_next(ast)
        finally:
            self._recurse_funs = c_prev_rf
            [setattr(self, *fnf) for fnf in c_restore]
            self._restore_stack.pop()
            if debug: print('leave uq/sot')

    # override these per dialect
    def eatom     (self, ast): return ast
    def atom      (self, ast): return ast
    def ekeyword  (self, ast): return ast
    def keyword   (self, ast): return ast
    def echarspec (self, ast): return ast
    def charspec  (self, ast): return ast
    def estring   (self, ast): return ast
    def string    (self, ast): return ast
    def quote     (self, ast): return ast
    def iquote    (self, ast): return ast
    def iq_uqot   (self, ast): return ast
    def iq_sunq   (self, ast): return ast
    def uquote    (self, ast): raise SyntaxError('unquote not in quasiquote or quote')
    # FIXME racket allows unquote in quote
    def q_uqot    (self, ast): return ast
    def q_sunq    (self, ast): return ast
    def suquote   (self, ast): raise SyntaxError('unquote splicing not in quasiquote')
    def listp     (self, ast): return List.from_ast(ast)  # pretty much everyone does this
    def lists     (self, ast): return ast
    def listc     (self, ast): return ast
    def sharp     (self, ast): return ast
    def sh_lstp   (self, ast): return self.listp(ast)
    def sh_lsts   (self, ast): return self.listp(ast)
    def sh_lstc   (self, ast): return self.listp(ast)
    def sh_eatm   (self, ast): return self.eatom(ast)
    def sh_atom   (self, ast): return self.atom(ast)
    def sh_ekey   (self, ast): return self.ekeyword(ast)
    def sh_keyw   (self, ast): return self.keyword(ast)
    def sh_estr   (self, ast): return self.estring(ast)
    def sh_strn   (self, ast): return self.string(ast)
    def sh_quot   (self, ast): return self.quote(ast)
    def sh_iqot   (self, ast): return self.iquote(ast)
    def sh_uqot   (self, ast): return self.uquote(ast)
    def sh_sunq   (self, ast): return self.suquote(ast)
    # TODO iq_sh_uqot


class WalkPl(Walk):

    def listp(self, ast):
        noc = [self(v) for v in ast.collect if not self._skip(v)]
        if all(isinstance(_, Keyword) for _ in noc[0::2]):
            # FIXME recurse ???
            # XXX indeed recursion can and does happen in Walk.__call__
            return PList(noc)
            #return PList.from_ast(ast)
        else:
            # we have to do this to handle cases where regular lists
            # are nested recursively inside, we catch "not a plist"
            # errors later
            return List(noc)
            #asdf = ast.collect[0::2]
            #breakpoint()
            #raise ValueError(f'wat\n{asdf}')

        return super().listp(ast)


# configure the parser

def configure(allow_in_symbol=tuple(),  # True, False, SyntaxError maybe dict for syntax error?
              additional_whitespace='',
              #escape_always=True,
              #curlies_map=False,
              #char_esc_expand=tuple(),  # expand rules must be provided at runtime FIXME dict or function?
              #char_auto_escape=True,  # FIXME the variety of behaviors has 2 axes I think, not clear what the default should be yet
              char_auto_end=True,  # FIXME same issue  XXX not sure we need this, somehow I think racket special cased these?
              #char_name_symbol=False,  # this is what cl does
              #immutable_cons=False,  # FIXME this should be deferred
              #sharp_atom_end_only_some=False,  # FIXME I think this is how to deal w/ racket?

              ## tokens (common are set)

              # the naming conventions for tokens
              # are to use beg end when a form must end with a specific token
              # and to use to for forms that have multiple possible ends

              t_newline='\n',
              t_tab='\t',
              t_space=' ',
              t_beg_list_p='(',
              t_end_list_p=')',
              t_beg_list_s=unp,
              t_end_list_s=unp,
              t_beg_list_c=unp,
              t_end_list_c=unp,
              t_beg_end_str='"',
              t_beg_end_aver=unp,
              t_to_sharp='#',
              t_to_cblk_in_shrp=unp,
              t_to_xcom_in_shrp=unp,
              t_to_hcom_in_shrp=unp,
              t_to_keyw_in_shrp=unp,
              t_to_feat_in_shrp=unp,
              t_to_strn_in_shrp=unp,  # TODO only used in hy, adds more states, not sure if want
              t_to_quote="'",
              t_to_quasi='`',
              t_to_unquote=',',
              t_to_unquote_in_quasi=unp,
              t_to_splc_in_unq='@',
              t_to_keyw=unp,
              t_to_esc=unp,
              t_to_esc_in_str='\\',
              t_to_char=unp,
              t_to_comment=';',

              **kwargs):

    eof = object()
    toks = [t_newline, t_tab, t_space,
            t_beg_list_p, t_end_list_p,
            t_beg_list_s, t_end_list_s,
            t_beg_list_c, t_end_list_c,
            t_beg_end_str, t_beg_end_aver,
            t_to_sharp, t_to_quote, t_to_quasi, t_to_unquote,
            # t_to_unquote_in_quasi, # FIXME not clear that this should be here XXX it can't
            t_to_keyw, t_to_esc, t_to_char, t_to_comment,]

    # even though additional whitespace doesn't get an explicit name
    # it still needs to be in syms to prevent it from being included
    # in identifier symbols
    toks.extend(additional_whitespace)

    # other symbols that we may want to support directly in the reader
    #am = '&'  # maybe want for lambda list control words?
    #qm = '?'
    #at = '@'  # TODO splicing unquote

    whitespace = (t_newline, t_tab, t_space) + tuple(_ for _ in additional_whitespace)
    scoe_complex = (t_newline, t_beg_end_str, t_to_esc, t_to_sharp, t_to_cblk_in_shrp)
    # note that pipe does not end atoms and keywords but the state still
    # transitions to v similarly exiting v does not exit atom or keyword
    ak_ends = (t_beg_list_p, t_end_list_p, t_beg_list_s, t_end_list_s, t_beg_list_c, t_end_list_c,
               t_to_quote, t_to_quasi, t_to_unquote, # t_to_unquote_in_quasi, # uiq can't be in due to context
               t_to_comment, t_beg_end_str,
               *whitespace)
    '()[]{};"\'`,\n\t '
    if allow_in_symbol:
        _ais = tuple(c for c in allow_in_symbol)  # sometimes unp is in ak_ends
        ak_ends = tuple(s for s in ak_ends if s not in _ais)

    m_ends = (*ak_ends, t_to_char, t_beg_end_aver, t_to_sharp)  # FIXME tricky because we need at least one char for m
    # m is s_char now it was question *m*ark before

    # XXX NOTE state could be defined globally but I think
    # keeping it in here will have better performance
    states = [bos, s_atom, s_keyw, s_quote, s_unquote, s_unq_first,
              s_unq_splice, s_quasiq, s_string, s_list_p, s_list_s,
              s_list_c, s_sharp, s_comm, s_esc, s_esc_str,
              s_comment_x, s_comment_h, s_feat_x, s_cblk, s_pi_in_cblk,
              s_atom_verbatim, s_comment_b_nest, s_char, s_char_first] = [
        State(i) for i in range(25)]
    esclikes = s_esc, s_char
    quotelikes = s_quote, s_unquote, s_unq_first, s_unq_splice, s_quasiq, s_comment_x, s_feat_x, s_sharp
    listlikes = s_list_p, s_list_s, s_list_c, s_cblk
    quasi_ignores = (s_comm, s_comment_x, s_comment_h, s_cblk, s_pi_in_cblk, s_comment_b_nest,
                     s_string, s_atom_verbatim, s_esc, s_esc_str, s_char
                     )  # XXX FIXME AAAAAAAAAAAAAAAAA
    #listlikes + (bos, s_atom, s_kewy)

    state_names = {
        bos: 'bos',
        s_atom: 'atom',
        s_keyw: 'keyw',
        s_quote: 'quot',
        s_unquote: 'unqu',
        s_unq_first: 'unqf',
        s_unq_splice: 'unqs',
        s_quasiq: 'qasi',
        s_string: 'strn',
        s_list_p: 'list',
        s_list_s: 'vect',
        s_list_c: 'dict',
        s_sharp: 'shrp',
        s_comm: 'comm',
        s_esc: 'escp',
        s_esc_str: 'sesc',
        s_comment_x: 'xcom',
        s_comment_h: 'hcom',
        s_feat_x: 'feat',
        s_cblk: 'cblk',
        s_pi_in_cblk: 'mebl',  # maybe leave block comment
        s_atom_verbatim: 'aver',
        s_comment_b_nest: 'nblk',
        s_char: 'char',
        s_char_first: 'chrf',
    }

    not_str = {
        unp: '<unparsable>',
        eof: '<end-of-file>',
        None: '<None>',
    }

    State.state_names.update(state_names)
    state_names = None
    del state_names

    _in_quasi = False  # XXX evil context sensative stack inspecting wormhole to insanity


    def caste_string(collect):
        try:
            return String(''.join(collect))  # FIXME syntax loc
        except TypeError:
            return EString(collect)

    def caste_atom(collect):
        return (EAtom(collect) if
                [_ for _ in collect if isinstance(_, Escape)] else
                Atom(collect))

    def caste_keyword(collect):
        return (EKeyword(collect) if
                [_ for _ in collect if isinstance(_, Escape)] else
                Keyword.from_collect(collect))

    def caste_sharp(collect):
        if collect:
            return Sharp.from_collect(collect)  # TODO override
        else:
            # defer error until later and
            # note that this is not an EAtom

            # returning an Atom here ensures that Sharp is never empty
            # and reduces the number of special cases that must be
            # handled there
            return Atom('#')

    def caste(state, collect):  # TODO start and end points
        # FIXME this second set of branches shouldn't be necessary?
        if   state == s_string: val = caste_string(collect)
        elif state == s_atom: val = caste_atom(collect)
        elif state == s_comm: val = Comment(collect)
        elif state == s_cblk: val = BComment(collect)
        elif state == s_keyw: val = caste_keyword(collect)
        elif state == s_list_p: val = ListP(collect)
        elif state == s_list_s: val = ListS(collect)
        elif state == s_list_c: val = ListC(collect)
        elif state == s_quote: val = Quote.from_collect(collect)
        elif state == s_quasiq: val = IQuote.from_collect(collect)
        elif state == s_unquote: val = UQuote.from_collect(collect)
        elif state == s_unq_splice: val = SUQuote.from_collect(collect)
        elif state == s_sharp: val = caste_sharp(collect)
        elif state == s_comment_x: val = XComment.from_collect(collect)
        elif state == s_comment_h: val = HComment.from_collect(collect)
        elif state == s_feat_x: val = FeatureExpr.from_collect(collect)  # TODO override
        elif state == s_atom_verbatim: val = caste_atom(collect)  # cl vs rr different behavior '|a \ b| '|a  b|
        elif state in (s_char_first, s_char): val = (
                ECharSpec(collect)
                if [_ for _ in collect if isinstance(_, SEscape)] else
                CharSpec(collect))
        # XXX if you hit this error with bos at eof DO NOT ADD bos here!
        else: raise UnknownStateError(f'Unknown state: {state} {collect}')
        # bind start and end points here for simplicity
        # val.start, val.end = start, end
        return val


    def process_collect(collect_stack, state, point_beg=None, point_end=None):
        thing = caste(state, collect_stack.pop())
        thing._point_beg = point_beg
        thing._point_end = point_end
        #print(thing, point_beg, point_end,
              #thing._point_beg, thing._point_end)
        return thing
        # XXXXXXXXXXX unused below
        thing = collect_stack.pop()
        if collect is not None:
            assert collect == thing, f'\n{collect}\n!=\n{thing}'
            thing = caste(state, collect)
        else:
            thing = caste(state, thing)  # nested case

        return thing


    def push_state_from_char(collect_stack, collect, stack, char):
        """ transition forward """
        nonlocal _in_quasi
        # worth the function call overhead for readability
        def to_state(next_state):
            stack.append(next_state)
            collect_stack.append([])
            return None

        def to_state_collect(next_state):
            stack.append(next_state)
            _collect = []
            collect_stack.append(_collect)
            return _collect

        if   char == t_beg_end_str: collect = to_state_collect(s_string)
        elif char == t_to_comment: collect = to_state_collect(s_comm)
        elif char == t_beg_list_p: collect = to_state(s_list_p)
        elif char == t_beg_list_s: collect = to_state(s_list_s)
        elif char == t_beg_list_c: collect = to_state(s_list_c)
        elif char == t_to_quote: collect = to_state(s_quote)
        elif char == t_to_quasi:
            collect = to_state(s_quasiq)
            _in_quasi = True
        elif char == t_to_unquote:
            collect = to_state(s_unq_first)
            _in_quasi = False
        elif char == t_to_unquote_in_quasi:
            collect = to_state(s_unq_first)
            _in_quasi = False
        elif char == t_beg_end_aver: collect = to_state_collect(s_atom_verbatim)
        # XXX warning: this branch for pipe only works correctly if
        # collect is None
        elif char == t_to_sharp:
            state = stack[-1]
            if state == s_cblk:  # FIXME do we really want to do this here
                # it means that we didn't handle a case correctly in parse
                stack.append(s_comment_b_nest)  # FIXME maybe nest s_cblk
            else:
                collect = to_state(s_sharp)
        elif char == t_to_char:
            collect = to_state_collect(s_char_first)
        elif char == t_to_esc_in_str:
            collect = to_state_collect(s_esc_str)
        else:
            raise Exception('This should never happen.')

        return collect


    def in_quasi(stack):  # XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
        # XXX FIXME WARNING I'm fairly certain that the call to `in
        # stack' here takes this parser from a pushdown automata to a
        # stack automata or at least something more powerful than a
        # pushdown automata, which moves us from context free to
        # context sensative which I think moves us into the dangerzone
        # of recursively enumerable languages which is a place that we
        # do not want to be XXX in principle it is possible to
        # implement this by duplicating the listlike states so that
        # they can be marked as being inside a quasiquote, however
        # then you get into major weirdness with nested quasiquotes
        # HOWEVER, given that in principle we could indeed implement this
        # just by adding states maybe it is ok to open this can of worms
        # since we know we are only using it reduce the number of states?
        return s_quasiq in stack and (
            s_unquote not in stack or
            # yeah, this is just as crazy as you think it is
            stack[::-1].index(s_unquote) > stack[::-1].index(s_quasiq))


    def resolve_pops(collect_stack, collect, stack, char, point_beg=None, point_end=None):
        state = stack[-1]
        *cut, = i_resolve_pops(collect_stack, stack, char, point_beg, point_end)

        if state == s_char_first:  # FIXME using this to prevent issues with strings sigh
            collect = None
        elif (char in (t_beg_list_p, t_beg_list_s, t_beg_list_c,
                       t_to_quote, t_to_quasi, t_to_unquote,
                       # TODO review to see if we are missing anything here
                       # given that we missed comment
                       t_to_comment) or
              state == s_char and char in (t_to_char, t_to_sharp) or  # and char_auto_end ??? chars may end chars ie not cl
              char == t_beg_end_str and state != s_string or
              char == t_to_unquote_in_quasi and in_quasi(stack)):  # FIXME WARNING in_quasi dangerzone
            collect = push_state_from_char(collect_stack, collect, stack, char)
        elif stack[-1] == s_cblk:
            collect = collect_stack[-1]
            # a bit different from other listlikes because it is its own collector
            # in addition to a nested container for other block quotes
        else:
            collect = None

        return (collect, *cut)


    def i_resolve_pops(collect_stack, stack, char, point_beg=None, point_end=None):
        if debug:
            print(f'RS: {char if isinstance(char, str) else not_str[char]} {stack}')
            print(f'CO: {collect_stack}')
            pass
        state_prev = stack.pop()  # pop ^
        if (state_prev in (s_pi_in_cblk,)):  # transition only no collect
            return i_resolve_pops(collect_stack, stack, char, point_beg, point_end)

        # FIXME #\;a causes isses for a bunch of readers which
        # this behavior avoids, but requires a special case
        #if state_prev == m and char == sc:

        cut = process_collect(collect_stack, state_prev, point_beg, point_end)  # cut /
        state = stack[-1]
        ##print(f'resp: {state_prev} {state} {cut!r}')

        if state_prev in listlikes and char in (t_end_list_p, t_end_list_s, t_end_list_c):
            char = None

        if char == eof:  # sigh using a branch for something that provably only happens once
            if state_prev in (s_string, s_atom_verbatim, s_char_first, *listlikes):
                # TODO unterminated x starting at position y

                raise SyntaxError(f'Unterminated thing {state_prev}.')

            if (state != bos or #and state_prev in (c, m, a) or
                state_prev == s_char_first):
                collect_stack[-1].append(cut)
                return i_resolve_pops(collect_stack, stack, char, point_beg, point_end)

            return cut,

        elif not collect_stack:
            return cut,
        elif state_prev in (s_comm, s_comment_h):
            if state in (*quotelikes, *listlikes):
                collect_stack[-1].append(cut)
                return tuple()
            else:
                return cut,
        elif state in quotelikes:
            # FIXME make sure features eat twice?
            collect_stack[-1].append(cut)
            if state_prev is s_comment_x and state is s_comment_x:
                # don't comment things that are already commented
                # comment the thing that comes after them
                # FIXME this gets very tricky with feature expressions
                return tuple()
            if state_prev == s_unquote:
                nonlocal _in_quasi
                _in_quasi = in_quasi(stack)  # FIXME SIGH EVIL SIGH

            return i_resolve_pops(collect_stack, stack, char, point_beg, point_end)
        elif (char in (t_end_list_p, t_end_list_s, t_end_list_c) and
              state_prev in (*quotelikes, s_atom, s_keyw, s_char, s_char_first)):
            # multiple things end at the same time
            collect_stack[-1].append(cut)
            if state_prev != s_char_first:
                return i_resolve_pops(collect_stack, stack, char, point_beg, point_end)
            else:
                return tuple()
        elif state in listlikes:
            collect_stack[-1].append(cut)
            return tuple()
        else:
            return cut,


    def process_h(collect_stack, stack, char):
        """ return True if should continue """
        # TODO in theory if we pass the generator in here
        # then we can make it possible to implement dispatch macros
        # transparently ...

        # pushed h, not popped h
        if char in (t_end_list_p, t_end_list_s, t_end_list_c): raise SyntaxError(f'# {char} not allowed.')
        elif char == t_to_cblk_in_shrp:  # FIXME essentially things that we want to ignore right now
            stack.pop()
            stack.append(s_cblk)
            # FIXME need to update collect
            #collect_stack.pop()
            return True
        elif char == t_to_esc:
            stack.pop()
            stack.append(s_char_first)
            return True
        elif char == t_to_xcom_in_shrp:
            stack.pop()
            # do not need collect_stack
            # reuse the collect from h for x
            stack.append(s_comment_x)
            return True
        elif char == t_to_hcom_in_shrp:  # NOTE this is like a line comment NOT like an x comment
            stack.pop()
            stack.append(s_comment_h)
            return True
        elif char == t_to_feat_in_shrp:  # FIXME
            stack.pop()
            # do not need collect_stack
            # reuse the collect from h for f
            stack.append(s_feat_x)
            return True


    def parse(iter_over_len_1_strings):
        """ First phase parse into a common representation
        of a generic lisp AST. A second, dialect specific
        phase is needed to complete the functionality needed
        for read. """
        bp = False
        if not isinstance(iter_over_len_1_strings, GeneratorType):
            iter_over_len_1_strings = (_ for _ in iter_over_len_1_strings)

        stack = [bos]
        collect_stack = []
        collect = None
        old_ps = point_start = None
        line = 1
        # FIXME also need a point stack at which point it is
        # likely better to switch to having a combined stack
        # for state, collect, and point
        last_line_point = 0  # to calculate column

        def mgen():
            nonlocal line, last_line_point
            # We use a generator function here so that if the
            # generator is passed to a dispatch macro form all the
            # book keeping machinery stays in sync.
            for point, char in enumerate(iter_over_len_1_strings):
                yield point, char
                if char == t_newline:
                    line += 1
                    last_line_point = point

        point = 0
        point_start_stack = [point]
        gen = mgen()
        in_quasi
        nonlocal _in_quasi
        _in_quasi = False
        try:
            for point, char in gen:
                state = stack[-1]

                if debug:
                    #print('IQ:', in_quasi(stack))
                    print('cs:', char, stack)
                    pass

                if state == s_unq_first:
                    # FIXME can we integrate this into the main switch?
                    if char == t_to_splc_in_unq:
                        stack.pop()
                        stack.append(s_unq_splice)
                        continue
                    else:
                        stack.pop()
                        stack.append(s_unquote)
                        state = s_unquote

                if state == s_sharp:
                    # make any transitions that are required,
                    # fail on failure cases, or proceed
                    # needs to come before escape due to #\
                    if process_h(collect_stack, stack, char):
                        if stack[-1] in (s_cblk, s_char_first):
                            collect = collect_stack[-1]  # FIXME h should be via push_state_from_char no?
                        continue

                # handle escape
                if state in (s_esc, s_esc_str):
                    # escape is only needed for a single char
                    stack.pop()
                    # FIXME TODO add explicit Ast node for this
                    # so we can differentiate between the raw forms?
                    # yes, this is better than the alternative which
                    # is to hardcode escape rules into the reader
                    if state == s_esc_str:
                        #char = caste(state, char)  # TODO make homogenous
                        char = SEscape(char)
                        char._point_beg = point
                        char._point_end = point
                        #breakpoint()
                    else:
                        char = Escape(char)
                        char._point_beg = point
                        char._point_end = point
                elif t_to_char == '?' and state in (s_char_first, s_char) and char == t_to_esc:  # FIXME HACK for elisp
                    stack.append(s_esc_str)  # NOTE elisp uses the same escape sequences for chars and strings
                    continue
                elif state not in (s_comm, s_comment_h, s_cblk, s_comment_b_nest, s_char_first, s_pi_in_cblk) and char == t_to_esc:  # implicit not in (c, o, m1)
                    # FIXME TODO need to split \ in string from the rest
                    if state in (s_string, s_char):
                        stack.append(s_esc_str)
                    else:
                        stack.append(s_esc)  # FIXME why do we do this? is it because of nested blk comments?
                    #breakpoint()  # XXX m debug start point
                    # FIXME interaction between m and e
                    continue
                elif state == s_string and char == t_to_esc_in_str:
                    # FIXME we should be able to handle this homogenously
                    stack.append(s_esc_str)
                    continue

                if state == s_char_first:
                    if char in toks and char in m_ends:
                        collect.append(char)
                    else:
                        # FIXME action at a distance
                        # this pops here to keep the scoe
                        # block a bit simpler
                        stack.pop()

                if state in (s_esc, s_esc_str) and stack[-1] == s_char_first: # state == m or
                    # FIXME m and e have sligh issues here
                    # needs to be handled after e
                    # ideally we could fall through
                    # since it pushes and pops on the same char
                    # but that means we would have to do something
                    # to reorder everything
                    # FIXME sigh, so now we copy and paste all the stuff from
                    # our usual resolve pops section here

                    collect = collect_stack[-1]  # inherit from the sharp or question mark?
                    collect.append(char)

                    if char not in toks or char not in m_ends:
                        stack.pop()
                        stack.append(s_char)

                    if char_auto_end and char in m_ends:
                        point_beg = point_start_stack.pop()  # FIXME may need to pop more times depending on how many things end at the same time XXX probably by passing the point stack
                        *thing, = i_resolve_pops(collect_stack, stack, char, point_beg, point)
                        collect = None
                        yield from thing
                        # FIXME not sure if correct
                        if thing:
                            old_ps = point_start  # for debug
                            point_start = None
                        if collect is not None:
                            point_start = point
                        point_start_stack.append(point)
                        continue
                    else:
                        continue   # FIXME ... not sure if this is right well it sorta works

                #_in_quasi = in_quasi(stack)  # FIXME hilariously inefficient
                if state == s_comment_b_nest:
                    stack.pop()
                    if char == t_to_cblk_in_shrp:
                        # FIXME sigh
                        stack.append(s_cblk)
                        collect = []
                        collect_stack.append(collect)
                    else:
                        collect.append(t_to_cblk_in_shrp)

                elif state == s_cblk and char == t_to_cblk_in_shrp:
                    stack.append(s_pi_in_cblk)

                # handle all the + cases except
                elif (((not _in_quasi)
                       # FIXME XXXXXXXX UGRHHRHR in_quasi COMPLEXITY
                       or char != t_to_unquote_in_quasi
                       # FIXME this is aweful, arbitrary, and
                       or _in_quasi and state in quasi_ignores) and
                      (char not in toks or
                       state in (s_atom, s_keyw) and char not in ak_ends and char != t_beg_end_aver or
                       state in (s_char_first, s_char) and char not in m_ends or
                       char == t_to_keyw or
                       state != s_unq_first and char == t_to_splc_in_unq or  # FIXME do we really need u1?
                       state == s_sharp and char in (t_to_sharp, t_to_keyw_in_shrp) or  # FIXME should probably work like an FeaureExpr?
                       state == s_atom_verbatim and char != t_beg_end_aver or
                       state == s_pi_in_cblk and char != t_to_sharp or  # fixme confusing
                       state in (s_string, s_comm, s_comment_h, s_cblk, s_esc, s_esc_str) and char not in scoe_complex or
                       state in (s_string, s_comm, s_comment_h,         s_esc, s_esc_str) and char in (t_to_sharp, t_to_cblk_in_shrp) or
                       state in (s_string,                      s_cblk, s_esc, s_esc_str) and char == t_newline or
                       state in (          s_comm, s_comment_h, s_cblk, s_esc, s_esc_str) and char in (t_beg_end_str, t_to_esc) or
                       (t_to_char == '?' and
                        # FIXME EEEEMMMMMAAAAAAAAAAAAACCCCCCSSSSSS
                       state == s_char and
                        char in m_ends and
                        (collect[-1] == escape_hat or
                         collect[-1] == '-' and
                         len(collect) > 1 and  # FIXME HACK to avoid index error should be a state
                         # implicitly checking if there is a KeyChar in collect
                         # TODO see if we accidentally induce bugs by reading
                         # things that the elisp reader would not
                     isinstance(collect [-2], SEscape))))):
                    if collect is None:  # if collect is a list it is
                        # for another state so safe to assign state
                        # here, also required to handle the case where
                        # the loop immediately terminates
                        state = s_keyw if char in (t_to_keyw, t_to_keyw_in_shrp) else s_atom
                        stack.append(state)
                        collect = []
                        collect_stack.append(collect)

                        if point_start is None:
                            point_start = point
                            point_start_stack.append(point)

                        if state == s_keyw:  # don't collect the colon that starts the keyword
                            continue

                    if state == s_pi_in_cblk and char != t_to_cblk_in_shrp:
                        stack.pop()
                        collect.append(t_to_cblk_in_shrp)
                    elif state == s_char_first:
                        stack.append(s_char)  # FIXME lurking transition

                    collect.append(char)
                    # FIXME if we are in state m and get here everything breaks
                    # this is where a goto would be really easy

                elif ((state in quotelikes or state == bos) and
                                            char in (t_end_list_p, t_end_list_s, t_end_list_c) or
                      state == s_list_p and char in (              t_end_list_s, t_end_list_c) or
                      state == s_list_s and char in (t_end_list_p,               t_end_list_c) or
                      state == s_list_c and char in (t_end_list_p, t_end_list_s)):
                    raise SyntaxError('No matching paren!')

                # majority of pops happen here
                elif (state in (s_atom, s_keyw) and (char in ak_ends or
                      (char == t_to_unquote_in_quasi and _in_quasi)) or  # XXX context sensative warning
                      state in (s_char_first, s_char) and char in m_ends or
                      state == s_list_p     and char == t_end_list_p or
                      state == s_list_s     and char == t_end_list_s or
                      state == s_list_c     and char == t_end_list_c or
                      state == s_string     and char == t_beg_end_str or
                      state in (s_comm, s_comment_h) and char == t_newline or
                      state == s_pi_in_cblk and char == t_to_sharp):
                    point_beg = point_start_stack.pop()
                    collect, *thing = resolve_pops(collect_stack, collect, stack, char, point_beg, point)
                    yield from thing
                    # FIXME not sure if correct
                    if thing:
                        old_ps = point_start  # for debug
                        point_start = None
                    if collect is not None:
                        point_start = point
                    point_start_stack.append(point)

                elif char == t_beg_end_aver:  # FIXME bad implementation should not need this special case
                    # pi and ha are cases where we might or might not want/need to updated collect
                    # which is why they are sort of separate because we stopped passing collect
                    # through push_state_from_char
                    if state == s_atom_verbatim:
                        stack.pop()
                        state = stack[-1]
                    else:
                        # yep, just as complex as expected :/
                        if state in listlikes or state in quotelikes or state in (s_atom, s_keyw, bos):
                            if collect is None:
                                stack.append(s_atom)  # : must come outside if you want a keyword
                                collect = push_state_from_char(collect_stack, collect, stack, char)
                            else:
                                stack.append(s_atom_verbatim)
                        if point_start is None:
                            point_start = point
                            point_start_stack.append(point)

                elif state == s_cblk and char == t_to_sharp:
                    push_state_from_char(collect_stack, collect, stack, char)

                elif (state in listlikes or
                      state in quotelikes or
                      state == bos):
                    if char in whitespace:
                        continue

                    collect = push_state_from_char(collect_stack, collect, stack, char)
                    if point_start is None:
                        point_start = point
                        point_start_stack.append(point)
                else:
                    raise Exception("I can't believe you've done this")

            if collect_stack:  # EOF case
                point_beg = point_start_stack.pop()
                collect, *thing = resolve_pops(collect_stack, collect, stack, eof, point_beg, point)
                yield from thing
                if stack and stack != [bos]:
                    # handle this special case out here to avoid
                    # repeatedly checking if we are at eof internally
                    raise Exception('NOT DONE YET')

        except Exception as err:
            if point_start is None:
                point_start = old_ps
            ps = point_start + 1 if point_start is not None else 0
            pe = point + 1 if point is not None else 0
            msg = f'error start {ps} end {pe} line {line}'
            if isinstance(err, UnicodeDecodeError):
                raise err  # position is already marked
            else:
                raise err.__class__(msg) from err


    return parse



""" sxpyr parse lisp

Usage:
    sxpyr parse [options] [<path>...]

Options:
    --fuzz
    -d --debug
"""

import pathlib
from pyontutils import clifun as clif
from sxpyr import sxpyr as sxpyrmod
from sxpyr._exports import *
from sxpyr.walks import WalkRkt, WalkCl, WalkEl
from sxpyr.sxpyr import (
    conf_read, Walk, conf_plist, WalkPl, plist_to_dict, PList, Ast,
    make_do_path,
    DispatchNotImplementedError)

parse_common = configure()
parse_plist = configure(**conf_plist)
parse_sxpyr = configure(**conf_sxpyr)
parse_union = configure(**conf_union)
parse_cl = configure(**conf_cl)
parse_el = configure(**conf_el)
parse_xel = configure(**conf_xel)
parse_rkt = configure(**conf_rkt)
parse_gui = configure(**conf_gui)
parse_clj = configure(**conf_clj)
parse_hy = configure(**conf_hy)
parse_trx = configure(**conf_trx)

read_rkt = conf_read(parse_rkt, WalkRkt)
read_el = conf_read(parse_el, WalkEl)


def readFromStdIn(stdin=None):
    from select import select
    if stdin is None:
        from sys import stdin
    if select([stdin], [], [], 0.0)[0]:
        return stdin


class Options(clif.Options):

    @property
    def path(self):
        return [pathlib.Path(path).expanduser() for path in self._args['<path>']]

    #@property
    #def stdin(self):
        #return self._args['-']


class Main(clif.Dispatcher):

    def default(self):
        raise NotImplementedError('oops')

    def parse(self):
        # TODO guess dialect
        sxpyrmod.debug = self.options.debug  # FIXME sigh naming imports etc

        parse = parse_rkt  # FIXME
        walk = WalkRkt()
        read = read_rkt  # FIXME
        def walk_ast(ast_it):
            # ah if only you could have names land like waterwheels on
            # a river, at each point without having to tuple
            for expression in ast_it:
                if self.options.debug:
                    print('read:', expression)
                if not walk._skip(expression):
                    transformed = walk(expression)
                    yield transformed

        if not self.options.path:
            stdin = readFromStdIn()
            source_gen = stdin,
        else:
            source_gen = self.options.path

        parse_path = make_do_path(parse)
        read_path = make_do_path(read)

        asts, wlks = [], []
        ast_fail, wlk_fail = [], []
        bad_walk  = []
        for path in source_gen:
            try:
                ast_generator = parse_path(path)
                ast = list(ast_generator)
                asts.append(ast)

                wlk_generator = (walk_ast(ast))
                wlk = list(wlk_generator)
                wlks.append(wlk)
                repr(wlk)
            except (SyntaxError,
                    DispatchNotImplementedError,
                    UnicodeDecodeError) as e:
                if len(asts) > len(wlks):
                    bad_walk.append(asts.pop())
                    wlk_fail.append((path, e))
                else:
                    ast_fail.append((path, e))
                
        #print(asts)
        print(wlks)
        print(bad_walk)
        if wlk_fail:
            print('walk failures', wlk_fail)
        if ast_fail:
            print('ast failures', ast_fail)
        return asts, wlks


def main():
    options, *ad = Options.setup(__doc__, version='sxpyr 0.0.0')

    main = Main(options)

    if main.options.debug:
        print(main.options)

    if options.fuzz:
        import os
        import sys
        import afl
        #afl.init()
        while afl.loop(55555):
            out = main()
            #sys.stdin.seek(0)

        os._exit(0)
    else:
        out = main()


if __name__ == '__main__':
    main()
